
import numpy as np
from collections.abc import Iterable
import h5py

import io

try:
    import sqlite3

    def convert_array(text):
        out = io.BytesIO(text)
        out.seek(0)
        return np.load(out)
    def adapt_array(arr):
        """
        http://stackoverflow.com/a/31312102/190597 (SoulNibbler)
        """
        out = io.BytesIO()
        np.save(out, arr)
        out.seek(0)
        return sqlite3.Binary(out.read())

    sqlite3.register_adapter(np.ndarray, adapt_array)
    sqlite3.register_converter("array", convert_array)
except ImportError:
    sqlite3 = None




class DBDataset:
    def __init__(self,dbfile,table="training",select_keys=None):
        if sqlite3 is None:
            raise ImportError("sqlite3 is not available")
        self.con = sqlite3.connect(dbfile, detect_types=sqlite3.PARSE_DECLTYPES)
        self.cur = self.con.cursor()
        self.table = table
        self.keys = [k[1] for k in self.cur.execute(f"PRAGMA table_info({self.table})")]
        if select_keys is not None:
            self.select_keys = [k for k in self.keys if k in select_keys]
        else:
            self.select_keys = self.keys
        self.N = self.cur.execute(f"SELECT MAX(rowid) FROM {self.table}").fetchone()[0]

    def __len__(self):
        return self.N
    
    def __getitem__(self,idx):
        if isinstance(idx,Iterable):
            indices = [i for i in idx]
        elif isinstance(idx,slice):
            indices = range(*idx.indices(self.N))
        else:
            indices = [idx]
        
        indices = [str(i+1) if i>=0 else str(i+self.N+1) for i in indices]
        query = f"SELECT {', '.join(self.select_keys)} FROM {self.table} WHERE rowid IN ({', '.join(indices)})"
        if len(indices)==1:
            data = self.cur.execute(query).fetchone()
            return {k:d for k,d in zip(self.select_keys,data)}
        else:
            return [{k:d for k,d in zip(self.select_keys,row)} for row in self.cur.execute(query)]

    def __iter__(self):
        for i in range(self.N):
            yield self[i]
    
    def __del__(self):
        self.con.close()
    
    def __repr__(self):
        return f"DBDataset({self.table})"
    
    def __str__(self):
        return f"DBDataset({self.table}) with {self.N} entries"
    
class H5Dataset:
    def __init__(self,h5file,table="training"):
        self.h5file = h5file
        self.table = table
        self.f = h5py.File(self.h5file,"r")
        self.dataset = self.f[self.table]
        self.N = len(self.dataset)
        self.keys = list(self.dataset["0"].keys())
    
    def __len__(self):
        return self.N
    
    def __getitem__(self,idx):
        if isinstance(idx,Iterable):
            indices = [i for i in idx]
        elif isinstance(idx,slice):
            indices = range(*idx.indices(self.N))
        else:
            indices = [idx]
        
        indices = [str(i) for i in indices]
        if len(indices)==1:
            data = self.dataset[indices[0]]
            return {k:data[k][()] for k in self.keys}
        else:
            return [{k:data[k][()] for k in self.keys} for data in self.dataset[indices]]
    
    def __iter__(self):
        for i in range(self.N):
            yield self[i]
    
    def __del__(self):
        self.f.close()


import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Optional, Union, List, Sequence, ClassVar, Dict
import math
import dataclasses
import numpy as np
from .nets import FullyConnectedNet
from ...utils import AtomicUnits as au
from functools import partial
from ...utils.periodic_table import (
    PERIODIC_TABLE_REV_IDX,
    PERIODIC_TABLE,
    EL_STRUCT,
    VALENCE_STRUCTURE,
    XENONPY_PROPS,
    SJS_COORDINATES,
    PERIODIC_COORDINATES,
    ATOMIC_IONIZATION_ENERGY,
    POLARIZABILITIES,
    D3_COV_RADII,
    VDW_RADII,
    OXIDATION_STATES,
)


class SpeciesEncoding(nn.Module):
    """A module that encodes chemical species information.

    FID: SPECIES_ENCODING
    """

    encoding: str = "random"
    """ The encoding to use. Can be one of "one_hot", "occupation", "electronic_structure", "properties", "sjs_coordinates", "random". 
        Multiple encodings can be concatenated using the "+" separator.
    """
    dim: int = 16
    """ The dimension of the encoding if not fixed by design."""
    zmax: int = 86
    """ The maximum atomic number to encode."""
    output_key: Optional[str] = None
    """ The key to use for the output in the returned dictionary."""

    species_order: Optional[Union[str, Sequence[str]]] = None
    """ The order of the species to use for the encoding. Only used for "onehot" encoding.
         If None, we encode all elements up to `zmax`."""
    trainable: bool = False
    """ Whether the encoding is trainable or fixed. Does not apply to "random" encoding which is always trainable."""
    extra_params: Dict = dataclasses.field(default_factory=dict)
    """ Dictionary of extra parameters for the basis."""

    FID: ClassVar[str] = "SPECIES_ENCODING"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:

        zmax = self.zmax
        if zmax <= 0 or zmax > len(PERIODIC_TABLE):
            zmax = len(PERIODIC_TABLE)

        zmaxpad = zmax + 2

        encoding = self.encoding.lower().strip()
        encodings = encoding.split("+")
        ############################
        conv_tensors = []

        if "one_hot" in encodings or "onehot" in encodings:
            if self.species_order is None:
                conv_tensor = np.eye(zmax)
                conv_tensor = np.concatenate(
                    [np.zeros((1, zmax)), conv_tensor, np.zeros((1, zmax))], axis=0
                )
            else:
                if isinstance(self.species_order, str):
                    species_order = [el.strip() for el in self.species_order.split(",")]
                else:
                    species_order = [el for el in self.species_order]
                conv_tensor = np.zeros((zmaxpad, len(species_order)))
                for i, s in enumerate(species_order):
                    conv_tensor[PERIODIC_TABLE_REV_IDX[s], i] = 1

            conv_tensors.append(conv_tensor)

        if "occupation" in encodings:
            conv_tensor = np.zeros((zmaxpad, (zmax + 1) // 2))
            for i in range(1, zmax + 1):
                conv_tensor[i, : i // 2] = 1
                if i % 2 == 1:
                    conv_tensor[i, i // 2] = 0.5

            conv_tensors.append(conv_tensor)

        if "electronic_structure" in encodings:
            Z = np.arange(1, zmax + 1).reshape(-1, 1)
            Zref = [zmax]
            e_struct = np.array(EL_STRUCT[1 : zmax + 1])
            eref = [2, 2, 6, 2, 6, 2, 10, 6, 2, 10, 6, 2, 14, 10, 6, 2, 14, 10, 6]
            v_struct = np.array(VALENCE_STRUCTURE[1 : zmax + 1])
            vref = [2, 6, 10, 14]
            if zmax <= 86:
                e_struct = e_struct[:, :15]
                eref = eref[:15]
            ref = np.array(Zref + eref + vref)
            conv_tensor = np.concatenate([Z, e_struct, v_struct], axis=1)
            conv_tensor = conv_tensor / ref[None, :]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )

            conv_tensors.append(conv_tensor)
        
        if "valence_structure" in encodings:
            vref = np.array([2, 6, 10, 14])
            conv_tensor = np.array(VALENCE_STRUCTURE[1 : zmax + 1])/vref[None,:]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)
        
        if "period_onehot" in encodings:
            periods = PERIODIC_COORDINATES[1 : zmax + 1, 0]-1
            conv_tensor = np.eye(7)[periods]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)
        
        if "period" in encodings:
            conv_tensor = (PERIODIC_COORDINATES[1 : zmax + 1, 0]/7.)[:,None]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)
        
        if "period_positional" in encodings:
            row = PERIODIC_COORDINATES[1 : zmax + 1, 0]
            drow = self.extra_params.get("drow", default=5)
            nrow = self.extra_params.get("nrow", default=100.0)

            conv_tensor = positional_encoding_static(row, drow, nrow)
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)


        if "properties" in encodings:
            props = np.array(XENONPY_PROPS)[1:-1]
            assert (
                self.zmax <= props.shape[0]
            ), f"zmax > {props.shape[0]} not supported for xenonpy properties"
            conv_tensor = props[1 : zmax + 1]
            mean = np.mean(props, axis=0)
            std = np.std(props, axis=0)
            conv_tensor = (conv_tensor - mean[None, :]) / std[None, :]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)
        
        
        if "valence_properties" in encodings:
            assert zmax <= 86, "Valence properties only available for zmax <= 86"
            Z = np.arange(1, zmax + 1).reshape(-1, 1)
            Zref = [zmax]
            Zinv = 1.0 / Z
            Zinvref = [1.0]
            v_struct = np.array(VALENCE_STRUCTURE[1 : zmax + 1])
            vref = [2, 6, 10, 14]
            ionization = np.array(ATOMIC_IONIZATION_ENERGY[1 : zmax + 1]).reshape(-1, 1)
            ionizationref = [0.5]
            polariz = np.array(POLARIZABILITIES[1 : zmax + 1]).reshape(-1, 1)
            polarizref = [np.median(polariz)]

            cov = np.array(D3_COV_RADII[1 : zmax + 1]).reshape(-1, 1)
            covref = [np.median(cov)]

            vdw = np.array(VDW_RADII[1 : zmax + 1]).reshape(-1, 1)
            vdwref = [np.median(vdw)]

            ref = np.array(
                Zref + Zinvref + ionizationref + polarizref + covref + vdwref + vref
            )
            conv_tensor = np.concatenate(
                [Z, Zinv, ionization, polariz, cov, vdw, v_struct], axis=1
            )

            conv_tensor = conv_tensor / ref[None, :]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)

        if "sjs_coordinates" in encodings:
            coords = np.array(SJS_COORDINATES)[1:-1]
            conv_tensor = coords[1 : zmax + 1]
            mean = np.mean(coords, axis=0)
            std = np.std(coords, axis=0)
            conv_tensor = (conv_tensor - mean[None, :]) / std[None, :]
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)

        if "positional" in encodings:
            coords = np.array(PERIODIC_COORDINATES)[1 : zmax + 1]
            row, col = coords[:, 0], coords[:, 1]
            drow = self.extra_params.get("drow", default=5)
            dcol = self.dim - drow
            nrow = self.extra_params.get("nrow", default=100.0)
            ncol = self.extra_params.get("ncol", default=1000.0)

            erow = positional_encoding_static(row, drow, nrow)
            ecol = positional_encoding_static(col, dcol, ncol)
            conv_tensor = np.concatenate([erow, ecol], axis=-1)
            dim = conv_tensor.shape[1]
            conv_tensor = np.concatenate(
                [np.zeros((1, dim)), conv_tensor, np.zeros((1, dim))], axis=0
            )
            conv_tensors.append(conv_tensor)

        if "oxidation" in encodings:
            states_set = sorted(set(sum(OXIDATION_STATES, [])))
            nstates = len(states_set)
            state_dict = {s: i for i, s in enumerate(states_set)}
            conv_tensor = np.zeros((zmaxpad, nstates))
            for i, states in enumerate(OXIDATION_STATES[1 : zmax + 1]):
                for s in states:
                    conv_tensor[i, state_dict[s]] = 1
            conv_tensors.append(conv_tensor)

        if len(conv_tensors) > 0:
            conv_tensor = np.concatenate(conv_tensors, axis=1)
            if self.trainable:
                conv_tensor = self.param(
                    "conv_tensor",
                    lambda key: jnp.asarray(conv_tensor, dtype=jnp.float32),
                )
            else:
                conv_tensor = jnp.asarray(conv_tensor, dtype=jnp.float32)
            conv_tensors = [conv_tensor]
        else:
            conv_tensors = []

        if "random" in encodings:
            rand_encoding = self.param(
                "rand_encoding",
                lambda key, shape: jax.nn.standardize(
                    jax.random.normal(key, shape, dtype=jnp.float32)
                ),
                (zmaxpad, self.dim),
            )
            conv_tensors.append(rand_encoding)
        
        if "zeros" in encodings:
            zeros_encoding = self.param(
                "zeros_encoding",
                lambda key, shape: jnp.zeros(shape, dtype=jnp.float32),
                (zmaxpad, self.dim),
            )
            conv_tensors.append(zeros_encoding)

        if "randint" in encodings:
            rand_encoding = self.param(
                "randint_encoding",
                lambda key, shape: jax.random.randint(key, shape, 0, 2).astype(
                    jnp.float32
                ),
                (zmaxpad, self.dim),
            )
            conv_tensors.append(rand_encoding)

        if "randtri" in encodings:
            rand_encoding = self.param(
                "randtri_encoding",
                lambda key, shape: (jax.random.randint(key, shape, 0, 3) - 1).astype(
                    jnp.float32
                ),
                (zmaxpad, self.dim),
            )
            conv_tensors.append(rand_encoding)

        assert len(conv_tensors) > 0, f"No encoding recognized in '{self.encoding}'"

        conv_tensor = jnp.concatenate(conv_tensors, axis=1)

        species = inputs["species"] if isinstance(inputs, dict) else inputs
        out = conv_tensor[species]
        ############################

        if isinstance(inputs, dict):
            output_key = self.name if self.output_key is None else self.output_key
            out = out.astype(inputs["coordinates"].dtype)
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class RadialBasis(nn.Module):
    """Computes a radial encoding of distances.

    FID: RADIAL_BASIS
    """

    end: float
    """ The maximum distance to consider."""
    start: float = 0.0
    """ The minimum distance to consider."""
    dim: int = 8
    """ The dimension of the basis."""
    graph_key: Optional[str] = None
    """ The key of the graph in the inputs."""
    output_key: Optional[str] = None
    """ The key to use for the output in the returned dictionary."""
    basis: str = "bessel"
    """ The basis to use. Can be one of "bessel", "gaussian", "gaussian_rinv", "fourier", "spooky"."""
    trainable: bool = False
    """ Whether the basis parameters are trainable or fixed."""
    enforce_positive: bool = False
    """ Whether to enforce distance-start to be positive"""
    gamma: float = 1.0 / (2 * au.BOHR)
    """ The gamma parameter for the "spooky" basis."""
    n_levels: int = 10
    """ The number of levels for the "levels" basis."""
    alt_bessel_norm: bool = False
    """ If True, use the (2/(end-start))**0.5 normalization for the bessel basis."""
    extra_params: Dict = dataclasses.field(default_factory=dict)
    """ Dictionary of extra parameters for the basis."""

    FID: ClassVar[str] = "RADIAL_BASIS"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        if self.graph_key is not None:
            x = inputs[self.graph_key]["distances"]
        else:
            x = inputs["distances"] if isinstance(inputs, dict) else inputs

        shape = x.shape
        x = x.reshape(-1)

        basis = self.basis.lower()
        ############################
        if basis == "bessel":
            c = self.end - self.start
            x = x[:, None] - self.start
            # if self.enforce_positive:
            #     x = jax.nn.softplus(x)

            if self.trainable:
                bessel_roots = self.param(
                    "bessel_roots",
                    lambda key, dim: jnp.asarray(
                        np.arange(1, dim + 1, dtype=x.dtype)[None, :] * (math.pi / c)
                    ),
                    self.dim,
                )
                norm = 1.0 / jnp.max(
                    bessel_roots
                )  # (2.0 / c) ** 0.5 /jnp.max(bessel_roots)
            else:
                bessel_roots = jnp.asarray(
                    np.arange(1, self.dim + 1, dtype=x.dtype)[None, :] * (math.pi / c)
                )
                norm = 1.0 / (
                    self.dim * math.pi / c
                )  # (2.0 / c) ** 0.5/(self.dim*math.pi/c)
            if self.alt_bessel_norm:
                norm = (2.0 / c) ** 0.5
            out = norm * jnp.sin(x * bessel_roots) / x

            if self.enforce_positive:
                out = jnp.where(x > 0, out * (1.0 - jnp.exp(-(x**2))), 0.0)

        elif basis == "gaussian":
            if self.trainable:
                roots = self.param(
                    "radial_centers",
                    lambda key, dim, start, end: jnp.linspace(
                        start, end, dim + 1, dtype=x.dtype
                    )[None, :-1],
                    self.dim,
                    self.start,
                    self.end,
                )
                eta = self.param(
                    "radial_etas",
                    lambda key, dim, start, end: jnp.full(
                        dim,
                        dim / (end - start),
                        dtype=x.dtype,
                    )[None, :],
                    self.dim,
                    self.start,
                    self.end,
                )

            else:
                roots = jnp.asarray(
                    np.linspace(self.start, self.end, self.dim + 1)[None, :-1],
                    dtype=x.dtype,
                )
                eta = jnp.asarray(
                    np.full(self.dim, self.dim / (self.end - self.start))[None, :],
                    dtype=x.dtype,
                )

            x = x[:, None]
            x2 = (eta * (x - roots)) ** 2
            out = jnp.exp(-x2)
            if self.enforce_positive:
                out = jnp.where(
                    x > self.start,
                    out * (1.0 - jnp.exp(-10 * (x - self.start) ** 2)),
                    0.0,
                )

        elif basis == "gaussian_rinv":
            rinv_high = 1.0 / max(self.start, 0.1)
            rinv_low = 1.0 / (0.8 * self.end)

            if self.trainable:
                roots = self.param(
                    "radial_centers",
                    lambda key, dim, rinv_low, rinv_high: jnp.linspace(
                        rinv_low, rinv_high, dim, dtype=x.dtype
                    )[None, :],
                    self.dim,
                    rinv_low,
                    rinv_high,
                )
                sigmas = self.param(
                    "radial_sigmas",
                    lambda key, dim, rinv_low, rinv_high: jnp.full(
                        dim,
                        2**0.5 / (2 * dim * rinv_low),
                        dtype=x.dtype,
                    )[None, :],
                    self.dim,
                    rinv_low,
                    rinv_high,
                )
            else:
                roots = jnp.asarray(
                    np.linspace(rinv_low, rinv_high, self.dim, dtype=x.dtype)[None, :]
                )
                sigmas = jnp.asarray(
                    np.full(
                        self.dim,
                        2**0.5 / (2 * self.dim * rinv_low),
                    )[None, :],
                    dtype=x.dtype,
                )

            rinv = 1.0 / x[:, None]

            out = jnp.exp(-((rinv - roots) ** 2) / sigmas**2)

        elif basis == "fourier":
            if self.trainable:
                roots = self.param(
                    "roots",
                    lambda key, dim: jnp.arange(dim, dtype=x.dtype)[None, :] * math.pi,
                    self.dim,
                )
            else:
                roots = jnp.asarray(
                    np.arange(self.dim)[None, :] * math.pi, dtype=x.dtype
                )
            c = self.end - self.start
            x = x[:, None] - self.start
            # if self.enforce_positive:
            #     x = jax.nn.softplus(x)
            norm = 1.0 / (0.25 + 0.5 * self.dim) ** 0.5
            out = norm * jnp.cos(x * roots / c)
            if self.enforce_positive:
                out = jnp.where(x > 0, out, norm)

        elif basis == "spooky":

            gamma = self.gamma
            if self.trainable:
                gamma = jnp.abs(
                    self.param("gamma", lambda key: jnp.asarray(gamma, dtype=x.dtype))
                )

            if self.enforce_positive:
                x = jnp.where(x - self.start > 1.0e-3, x - self.start, 1.0e-3)[:, None]
                dim = self.dim
            else:
                x = x[:, None] - self.start
                dim = self.dim - 1

            norms = []
            for k in range(self.dim):
                norms.append(math.comb(dim, k))
            norms = jnp.asarray(np.array(norms)[None, :], dtype=x.dtype)

            e = jnp.exp(-gamma * x)
            k = jnp.asarray(np.arange(self.dim, dtype=x.dtype)[None, :])
            b = e**k * (1 - e) ** (dim - k)
            out = b * e * norms
            if self.enforce_positive:
                out = jnp.where(x > 1.0e-3, out * (1.0 - jnp.exp(-(x**2))), 0.0)
            # logfac = np.zeros(self.dim)
            # for i in range(2, self.dim):
            #     logfac[i] = logfac[i - 1] + np.log(i)
            # k = np.arange(self.dim)
            # n = self.dim - 1 - k
            # logbin = jnp.asarray((logfac[-1] - logfac[k] - logfac[n])[None,:], dtype=x.dtype)
            # n = jnp.asarray(n[None,:], dtype=x.dtype)
            # k = jnp.asarray(k[None,:], dtype=x.dtype)

            # gamma = 1.0 / (2 * au.BOHR)
            # if self.trainable:
            #     gamma = jnp.abs(
            #         self.param("gamma", lambda key: jnp.asarray(gamma, dtype=x.dtype))
            #     )
            # gammar = (-gamma * x)[:,None]
            # x = logbin + n * gammar + k * jnp.log(-jnp.expm1(gammar))
            # out = jnp.exp(x)*jnp.exp(gammar)
        elif basis == "levels":
            assert self.n_levels >= 2, "Number of levels must be >= 2."

            def initialize_levels(key):
                key0, key1, key_phi = jax.random.split(key, 3)
                level0 = jax.random.randint(key0, (self.dim,), 0, 2)
                level1 = jax.random.randint(key1, (self.dim,), 0, 2)
                # level0 = jax.random.normal(key0, (self.dim,), dtype=jnp.float32)
                # level1 = jax.random.normal(key1, (self.dim,), dtype=jnp.float32)
                phi = jax.random.uniform(key_phi, (self.dim,), dtype=jnp.float32)
                levels = [level0]
                for l in range(2, self.n_levels - 1):
                    tau = float(self.n_levels - l) / float(self.n_levels - 1)
                    phil = phi < tau
                    level = jnp.where(phil, level0, level1)
                    levels.append(level)
                levels.append(level1)
                return jnp.stack(levels).astype(jnp.float32)

            levels = self.param("levels", initialize_levels)
            # levels = self.param(
            #     "levels",
            #     lambda key, shape: jax.random.normal(key, shape, dtype=jnp.float32),
            #     (self.n_levels,self.dim),
            # )

            flevel = (x - self.start) / (self.end - self.start) * (self.n_levels - 1)
            ilevel = jnp.floor(flevel).astype(jnp.int32)
            ilevel1 = jnp.clip(ilevel + 1, 0, self.n_levels - 1)
            ilevel = jnp.clip(ilevel, 0, self.n_levels - 1)

            dx = flevel - ilevel
            w = 0.5 * (1 + jnp.cos(jnp.pi * dx))[:, None]

            ## interpolate between level vectors
            v1 = levels[ilevel]
            v2 = levels[ilevel1]
            out = v1 * w + v2 * (1 - w)
        elif basis == "finite_support":
            flevel = (x - self.start) / (self.end - self.start) * (self.dim + 1)
            ilevel = jnp.floor(flevel).astype(jnp.int32)
            ilevel1 = jnp.clip(ilevel + 1, 0, self.dim + 1)
            ilevel = jnp.clip(ilevel, 0, self.dim + 1)

            dx = flevel - ilevel
            w = 0.5 * (1 + jnp.cos(jnp.pi * dx))

            ilevelflat = ilevel + jnp.arange(x.shape[0]) * (self.dim + 2)
            ilevel1flat = ilevel1 + jnp.arange(x.shape[0]) * (self.dim + 2)

            out = (
                jnp.zeros((x.shape[0] * (self.dim + 2)), dtype=x.dtype)
                .at[ilevelflat]
                .set(w)
                .at[ilevel1flat]
                .set(1 - w)
                .reshape(-1, self.dim + 2)[:, 1:-1]
            )

        elif basis == "exp_lr" or basis == "exp":
            zeta = self.extra_params.get("zeta", default=2.0)
            s = self.extra_params.get("s", default=0.5)
            n = np.arange(self.dim)
            # if self.trainable:
            # zeta = jnp.abs(
            #     self.param("zeta", lambda key: jnp.asarray(zeta, dtype=x.dtype))
            # )
            # s = jnp.abs(self.param("s", lambda key: jnp.asarray(s, dtype=x.dtype)))

            a = zeta * s**n
            xx = np.linspace(0, self.end, 10000)

            if self.start > 0:
                a1 = np.minimum(1.0 / a, 1.0)
                switchstart = jnp.where(
                    x[:, None] < self.end,
                    1
                    - (
                        0.5
                        + 0.5
                        * jnp.cos(
                            np.pi * (x[:, None] - self.start) / (self.end - self.start)
                        )
                    )
                    ** a1[None, :],
                    1,
                ) * (x[:, None] > self.start)
                switchstartxx = (
                    1
                    - (
                        0.5
                        + 0.5
                        * np.cos(
                            np.pi * (xx[:, None] - self.start) / (self.end - self.start)
                        )
                    )
                    ** a1[None, :]
                ) * (xx[:, None] > self.start)

            else:
                switchstart = 1.0
                switchstartxx = 1.0

            norm = 1.0 / np.trapz(
                switchstartxx * np.exp(-a[None, :] * xx[:, None]), xx, axis=0
            )
            # norm = 1./np.max(switchstartxx*np.exp(-a[None, :] * xx[:, None]), axis=0)

            if self.trainable:
                a = jnp.abs(self.param("exponent", lambda key: jnp.asarray(a)))

            out = switchstart * jnp.exp(-a[None, :] * x[:, None]) * norm[None, :]

        elif basis == "neural_net" or basis == "nn":
            neurons = [*self.extra_params.get(
                "hidden_neurons", default=[2 * self.dim]
            ), self.dim]
            activation = self.extra_params.get("activation", default="swish")
            use_bias = self.extra_params.get("use_bias", default=True)

            out = FullyConnectedNet(
                neurons, activation=activation, use_bias=use_bias, squeeze=False
            )(x[:, None])
        elif basis == "damped_coulomb":
            l = np.arange(self.dim)[None, :]
            a = self.extra_params.get("a", default=1.0)
            if self.trainable:
                a = self.param("a", lambda key: jnp.asarray([a]*self.dim)[None,:])
            a2=a**2
            x = x[:, None] - self.start
            end = self.end - self.start
            x21 = a2 + x**2
            R21 = a2 + end**2
            out = (
                1.0 / x21 ** (0.5 * (l + 1))
                - 1.0 / (R21 ** (0.5 * (l + 1)))
                + (x - end) * ((l + 1) * end / (R21 ** (0.5 * (l + 3))))
            ) * (x < end)

        elif basis.startswith("spherical_bessel_"):
            l = int(basis.split("_")[-1])
            out = generate_spherical_jn_basis(self.dim, self.end,[l])(x)
        else:
            raise NotImplementedError(f"Unknown radial basis {basis}.")
        ############################

        out = out.reshape((*shape, self.dim))

        if self.graph_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out}
        return out


def positional_encoding_static(t, d: int, n: float = 10000.0):
    if d % 2 == 0:
        k = np.arange(d // 2)
    else:
        k = np.arange((d + 1) // 2)
    wk = np.asarray(1.0 / (n ** (2 * k / d)))
    wkt = wk[None, :] * t[:, None]
    out = np.concatenate([np.cos(wkt), np.sin(wkt)], axis=-1)
    if d % 2 == 1:
        out = out[:, :-1]
    return out


@partial(jax.jit, static_argnums=(1, 2), inline=True)
def positional_encoding(t, d: int, n: float = 10000.0):
    if d % 2 == 0:
        k = np.arange(d // 2)
    else:
        k = np.arange((d + 1) // 2)
    wk = jnp.asarray(1.0 / (n ** (2 * k / d)))
    wkt = wk[None, :] * t[:, None]
    out = jnp.concatenate([jnp.cos(wkt), jnp.sin(wkt)], axis=-1)
    if d % 2 == 1:
        out = out[:, :-1]
    return out


def generate_spherical_jn_basis(dim:int, rc:float, ls:Union[int,Sequence[int]]=[0], print_code:bool=False, jit:bool=False):
    from sympy import Symbol, jn, expand_func
    from scipy.special import spherical_jn
    from sympy import jn_zeros
    import scipy.integrate as integrate

    if isinstance(ls, int):
        ls = list(range(ls + 1))
    zl = [Symbol(f"xz[...,{l}]") for l in ls]
    zn = np.array([jn_zeros(l, dim) for l in ls], dtype=float).T
    znrc = zn / rc
    norms = np.zeros((dim, len(ls)), dtype=float)
    for il,l in enumerate(ls):
        for i in range(dim):
            norms[i, il] = (
                integrate.quad(lambda x: (spherical_jn(l, x) * x) ** 2, 0, zn[i, il])[0]
                / znrc[i, il] ** 3
            ) ** (-0.5)

    fn_str = f"""def spherical_jn_basis_(x):
    from jax.numpy import cos,sin
    
    znrc = jnp.array({znrc.tolist()},dtype=x.dtype)
    norms = jnp.array({norms.tolist()},dtype=x.dtype)
    xshape = x.shape
    x = x.reshape(-1)
    xz = x[:,None,None]*znrc[None,:,:]

    jns = jnp.stack([
  """
    for il,l in enumerate(ls):
        fn_str += f"      {expand_func(jn(l, zl[il]))},\n"
    fn_str += f"""    ],axis=-1)
    return (norms[None,:,:]*jns).reshape(*xshape,{dim},{len(ls)})
  """

    if print_code:
        print(fn_str)
    exec(fn_str)
    jn_basis = locals()["spherical_jn_basis_"]
    if jit:
        jn_basis = jax.jit(jn_basis)
    return jn_basis



import flax.linen as nn
from typing import Sequence, Callable, Union, Optional, ClassVar, Tuple
from ...utils.periodic_table import PERIODIC_TABLE_REV_IDX, PERIODIC_TABLE
import jax.numpy as jnp
import jax
import numpy as np
from functools import partial
from ...utils.activations import activation_from_str, TrainableSiLU
from ...utils.initializers import initializer_from_str, scaled_orthogonal
from flax.core import FrozenDict


class FullyConnectedNet(nn.Module):
    """A fully connected neural network module.

    FID: NEURAL_NET
    """

    neurons: Sequence[int]
    """A sequence of integers representing the dimensions of the network."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use."""
    use_bias: bool = True
    """Whether to use bias in the dense layers."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use."""

    FID: ClassVar[str] = "NEURAL_NET"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        """Applies the neural network to the given inputs.

        Args:
            inputs (Union[dict, jax.Array]): If input_key is None, a JAX array containing the inputs to the neural network. Else, a dictionary containing the inputs at the key input_key.

        Returns:
            Union[dict, jax.Array]: If output_key is None, the output tensor of the neural network. Else, a dictionary containing the original inputs and the output tensor at the key output_key.
        """
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        # activation = (
        #     activation_from_str(self.activation)
        #     if isinstance(self.activation, str)
        #     else self.activation
        # )
        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )
        ############################
        if isinstance(self.activation, str) and self.activation.lower() == "swiglu":
            for i, d in enumerate(self.neurons[:-1]):
                y = nn.Dense(
                    d,
                    use_bias=self.use_bias,
                    name=f"Layer_{i+1}",
                    kernel_init=kernel_init,
                )(x)
                z = jax.nn.swish(nn.Dense(
                    d,
                    use_bias=self.use_bias,
                    name=f"Mask_{i+1}",
                    kernel_init=kernel_init,
                )(x))
                x = y * z
        else:
            for i, d in enumerate(self.neurons[:-1]):
                x = nn.Dense(
                    d,
                    use_bias=self.use_bias,
                    name=f"Layer_{i+1}",
                    kernel_init=kernel_init,
                )(x)
                x = activation_from_str(self.activation)(x)
        x = nn.Dense(
            self.neurons[-1],
            use_bias=self.use_bias,
            name=f"Layer_{len(self.neurons)}",
            kernel_init=kernel_init,
        )(x)
        if self.squeeze and x.shape[-1] == 1:
            x = jnp.squeeze(x, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: x} if output_key is not None else x
        return x


class ResMLP(nn.Module):
    """Residual neural network as defined in the SpookyNet paper.
    
    FID: RES_MLP
    """

    use_bias: bool = True
    """Whether to include bias in the linear layers."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""

    kernel_init: Union[str, Callable] = "scaled_orthogonal(mode='fan_avg')"
    """The kernel initialization method to use."""
    res_only: bool = False
    """Whether to only apply the residual connection without additional activation and linear layer."""

    FID: ClassVar[str] = "RES_MLP"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )
        ############################
        out = nn.Dense(x.shape[-1], use_bias=self.use_bias, kernel_init=kernel_init)(
            TrainableSiLU()(x)
        )
        out = x + nn.Dense(
            x.shape[-1], use_bias=self.use_bias, kernel_init=nn.initializers.zeros
        )(TrainableSiLU()(out))

        if not self.res_only:
            out = nn.Dense(
                x.shape[-1], use_bias=self.use_bias, kernel_init=kernel_init
            )(TrainableSiLU()(out))
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class FullyResidualNet(nn.Module):
    """A neural network with skip connections at each layer.
    
    FID: SKIP_NET
    """

    dim: int
    """The dimension of the hidden layers."""
    output_dim: int
    """The dimension of the output layer."""
    nlayers: int
    """The number of layers in the network."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use."""
    use_bias: bool = True
    """Whether to include bias terms in the linear layers."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use."""

    FID: ClassVar[str] = "SKIP_NET"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        # activation = (
        #     activation_from_str(self.activation)
        #     if isinstance(self.activation, str)
        #     else self.activation
        # )
        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )
        ############################
        if x.shape[-1] != self.dim:
            x = nn.Dense(
                self.dim,
                use_bias=self.use_bias,
                name=f"Reshape",
                kernel_init=kernel_init,
            )(x)

        for i in range(self.nlayers - 1):
            x = x + activation_from_str(self.activation)(
                nn.Dense(
                    self.dim,
                    use_bias=self.use_bias,
                    name=f"Layer_{i+1}",
                    kernel_init=kernel_init,
                )(x)
            )
        x = nn.Dense(
            self.output_dim,
            use_bias=self.use_bias,
            name=f"Layer_{self.nlayers}",
            kernel_init=kernel_init,
        )(x)
        if self.squeeze and x.shape[-1] == 1:
            x = jnp.squeeze(x, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: x} if output_key is not None else x
        return x


class HierarchicalNet(nn.Module):
    """Neural network for a sequence of inputs (in axis=-2) with a decay factor
    
    FID: HIERARCHICAL_NET
    """

    neurons: Sequence[int]
    """A sequence of integers representing the number of neurons in each layer."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use."""
    use_bias: bool = True
    """Whether to include bias terms in the linear layers."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    decay: float = 0.01
    """The decay factor to scale each element of the sequence."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use."""

    FID: ClassVar[str] = "HIERARCHICAL_NET"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        ############################
        networks = nn.vmap(
            FullyConnectedNet,
            variable_axes={"params": 0},
            split_rngs={"params": True},
            in_axes=-2,
            out_axes=-2,
            kenel_init=self.kernel_init,
        )(self.neurons, self.activation, self.use_bias)

        out = networks(x)
        # scale each layer by a decay factor
        decay = jnp.asarray([self.decay**i for i in range(out.shape[-2])])
        out = out * decay[..., :, None]

        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class SpeciesIndexNet(nn.Module):
    """Chemical-species-specific neural network using precomputed species index.

    FID: SPECIES_INDEX_NET

    A neural network that applies a species-specific fully connected network to each atom embedding.
    A species index must be provided to filter the embeddings for each species and apply the corresponding network.
    This index can be obtained using the SPECIES_INDEXER preprocessing module from `fennol.models.preprocessing.SpeciesIndexer`

    """

    output_dim: int
    """The dimension of the output of the fully connected networks."""
    hidden_neurons: Union[dict, FrozenDict, Sequence[int]]
    """The hidden dimensions of the fully connected networks.
        If a dictionary is provided, it should map species names to dimensions.
        If a sequence is provided, the same dimensions will be used for all species."""
    species_order: Optional[Union[str, Sequence[str]]] = None
    """The species for which to build a network. Only required if neurons is not a dictionary."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use in the fully connected networks."""
    use_bias: bool = True
    """Whether to include bias terms in the fully connected networks."""
    input_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the embeddings of the atoms."""
    species_index_key: str = "species_index"
    """The key in the input dictionary that corresponds to the species index of the atoms. See `fennol.models.preprocessing.SpeciesIndexer`"""
    output_key: Optional[str] = None
    """The key in the output dictionary that corresponds to the network's output."""

    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method for the fully connected networks."""
    check_unhandled: bool = True

    FID: ClassVar[str] = "SPECIES_INDEX_NET"

    def setup(self):
        if not (
            isinstance(self.hidden_neurons, dict)
            or isinstance(self.hidden_neurons, FrozenDict)
        ):
            assert (
                self.species_order is not None
            ), "species_order must be provided if hidden_neurons is not a dictionary"
            if isinstance(self.species_order, str):
                species_order = [el.strip() for el in self.species_order.split(",")]
            else:
                species_order = [el for el in self.species_order]
            neurons = {k: self.hidden_neurons for k in species_order}
        else:
            neurons = self.hidden_neurons
            species_order = list(neurons.keys())
        for species in species_order:
            assert (
                species in PERIODIC_TABLE
            ), f"species {species} not found in periodic table"

        self.networks = {
            k: FullyConnectedNet(
                [*neurons[k], self.output_dim],
                self.activation,
                self.use_bias,
                name=k,
                kernel_init=self.kernel_init,
            )
            for k in species_order
        }

    def __call__(
        self, inputs: Union[dict, Tuple[jax.Array, jax.Array]]
    ) -> Union[dict, jax.Array]:

        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            species, embedding, species_index = inputs
        else:
            species, embedding = inputs["species"], inputs[self.input_key]
            species_index = inputs[self.species_index_key]

        assert isinstance(
            species_index, dict
        ), "species_index must be a dictionary for SpeciesIndexNetHet"

        ############################
        # initialization => instantiate all networks
        if self.is_initializing():
            x = jnp.zeros((1, embedding.shape[-1]), dtype=embedding.dtype)
            [net(x) for net in self.networks.values()]

        if self.check_unhandled:
            for b in species_index.keys():
                if b not in self.networks.keys():
                    raise ValueError(f"Species {b} not found in networks. Handled species are {self.networks.keys()}")

        ############################
        outputs = []
        indices = []
        for s, net in self.networks.items():
            if s not in species_index:
                continue
            idx = species_index[s]
            o = net(embedding[idx])
            outputs.append(o)
            indices.append(idx)

        o = jnp.concatenate(outputs, axis=0)
        idx = jnp.concatenate(indices, axis=0)

        out = (
            jnp.zeros((species.shape[0], *o.shape[1:]), dtype=o.dtype)
            .at[idx]
            .set(o, mode="drop")
        )

        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class ChemicalNet(nn.Module):
    """optimized Chemical-species-specific neural network.

    FID: CHEMICAL_NET

    A neural network that applies a fully connected network to each atom embedding in a chemical system and selects the output corresponding to the atom's species.
    This is an optimized version of ChemicalNetHet that uses vmap to apply the networks to all atoms at once.
    The optimization is allowed because all networks have the same shape.

    """

    species_order: Union[str, Sequence[str]]
    """The species for which to build a network."""
    neurons: Sequence[int]
    """The dimensions of the fully connected networks."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use in the fully connected networks."""
    use_bias: bool = True
    """Whether to include bias terms in the fully connected networks."""
    input_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the embeddings of the atoms."""
    output_key: Optional[str] = None
    """The key in the output dictionary that corresponds to the network's output."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method for the fully connected networks."""

    FID: ClassVar[str] = "CHEMICAL_NET"

    @nn.compact
    def __call__(
        self, inputs: Union[dict, Tuple[jax.Array, jax.Array]]
    ) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            species, embedding = inputs
        else:
            species, embedding = inputs["species"], inputs[self.input_key]

        ############################
        # build species to network index mapping (static => fixed when jitted)
        rev_idx = PERIODIC_TABLE_REV_IDX
        maxidx = max(rev_idx.values())
        if isinstance(self.species_order, str):
            species_order = [el.strip() for el in self.species_order.split(",")]
        else:
            species_order = [el for el in self.species_order]
        nspecies = len(species_order)
        conv_tensor_ = np.full((maxidx + 2,), -1, dtype=np.int32)
        for i, s in enumerate(species_order):
            conv_tensor_[rev_idx[s]] = i
        conv_tensor = jnp.asarray(conv_tensor_)
        indices = conv_tensor[species]

        ############################
        # build shape-sharing networks using vmap
        networks = nn.vmap(
            FullyConnectedNet,
            variable_axes={"params": 0},
            split_rngs={"params": True},
            in_axes=0,
        )(self.neurons, self.activation, self.use_bias, kernel_init=self.kernel_init)
        # repeat input along a new axis to compute for all species at once
        x = jnp.broadcast_to(
            embedding[None, :, :], (nspecies, *embedding.shape)
        )

        # apply networks to input and select the output corresponding to the species
        out = jnp.squeeze(
            jnp.take_along_axis(networks(x), indices[None, :, None], axis=0), axis=0
        )

        out = jnp.where((indices >= 0)[:, None], out, 0.0)
        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class MOENet(nn.Module):
    """Mixture of Experts neural network.

    FID: MOE_NET

    This class represents a Mixture of Experts neural network. It takes in an input and applies a set of shape-sharing networks
    to the input based on a router. The outputs of the shape-sharing networks are then combined using weights computed by the router.

    """

    neurons: Sequence[int]
    """A sequence of integers representing the number of neurons in each shape-sharing network."""
    num_networks: int
    """The number of shape-sharing networks to create."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use in the shape-sharing networks."""
    use_bias: bool = True
    """Whether to include bias in the shape-sharing networks."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use in the shape-sharing networks."""
    router_key: Optional[str] = None
    """The key of the router tensor. If None, the router is assumed to be the same as the input tensor."""

    FID: ClassVar[str] = "MOE_NET"

    @nn.compact
    def __call__(
        self, inputs: Union[dict, Tuple[jax.Array, jax.Array]]
    ) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            if isinstance(inputs, tuple):
                embedding, router = inputs
            else:
                embedding = router = inputs
        else:
            embedding = inputs[self.input_key]
            router = (
                inputs[self.router_key] if self.router_key is not None else embedding
            )

        ############################
        # build shape-sharing networks using vmap
        networks = nn.vmap(
            FullyConnectedNet,
            variable_axes={"params": 0},
            split_rngs={"params": True},
            in_axes=0,
            out_axes=0,
        )(self.neurons, self.activation, self.use_bias, kernel_init=self.kernel_init)
        # repeat input along a new axis to compute for all networks at once
        x = jnp.repeat(embedding[None, :, :], self.num_networks, axis=0)

        w = nn.softmax(nn.Dense(self.num_networks, name="router")(router), axis=-1)

        out = (networks(x) * w.T[:, :, None]).sum(axis=0)

        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out

class ChannelNet(nn.Module):
    """Apply a different neural network to each channel.
    
    FID: CHANNEL_NET
    """

    neurons: Sequence[int]
    """A sequence of integers representing the number of neurons in each shape-sharing network."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use in the shape-sharing networks."""
    use_bias: bool = True
    """Whether to include bias in the shape-sharing networks."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use in the shape-sharing networks."""
    channel_axis: int = -2
    """The axis to use as channel. Its length will be the number of shape-sharing networks."""

    FID: ClassVar[str] = "CHANNEL_NET"

    @nn.compact
    def __call__(
        self, inputs: Union[dict, Tuple[jax.Array, jax.Array]]
    ) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        ############################
        # build shape-sharing networks using vmap
        networks = nn.vmap(
            FullyConnectedNet,
            variable_axes={"params": 0},
            split_rngs={"params": True},
            in_axes=self.channel_axis,
            out_axes=self.channel_axis,
        )(self.neurons, self.activation, self.use_bias, kernel_init=self.kernel_init)

        out = networks(x)

        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


class GatedPerceptron(nn.Module):
    """Gated Perceptron neural network.

    FID: GATED_PERCEPTRON

    This class represents a Gated Perceptron neural network model. It applies a gating mechanism
    to the input data and performs linear transformation using a dense layer followed by an activation function.
    """

    dim: int
    """The dimensionality of the output space."""
    use_bias: bool = True
    """Whether to include a bias term in the dense layer."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use."""

    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""

    FID: ClassVar[str] = "GATED_PERCEPTRON"

    @nn.compact
    def __call__(self, inputs):
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            x = inputs
        else:
            x = inputs[self.input_key]

        # activation = (
        #     activation_from_str(self.activation)
        #     if isinstance(self.activation, str)
        #     else self.activation
        # )
        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )
        ############################
        gate = jax.nn.sigmoid(
            nn.Dense(self.dim, use_bias=self.use_bias, kernel_init=kernel_init)(x)
        )
        x = gate * activation_from_str(self.activation)(
            nn.Dense(self.dim, use_bias=self.use_bias, kernel_init=kernel_init)(x)
        )

        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: x} if output_key is not None else x
        return x


class ZAcNet(nn.Module):
    """ A fully connected neural network module with affine Z-dependent adjustments of activations.
    
    FID: ZACNET
    """

    neurons: Sequence[int]
    """A sequence of integers representing the dimensions of the network."""
    zmax: int = 86
    """The maximum atomic number to consider."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use."""
    use_bias: bool = True
    """Whether to use bias in the dense layers."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use."""
    species_key: str = "species"
    """The key of the species tensor."""

    FID: ClassVar[str] = "ZACNET"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            species, x = inputs
        else:
            species, x = inputs[self.species_key], inputs[self.input_key]

        # activation = (
        #     activation_from_str(self.activation)
        #     if isinstance(self.activation, str)
        #     else self.activation
        # )
        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )
        ############################
        for i, d in enumerate(self.neurons[:-1]):
            x = nn.Dense(
                d, use_bias=self.use_bias, name=f"Layer_{i+1}", kernel_init=kernel_init
            )(x)
            sig = self.param(
                f"sig_{i+1}",
                lambda key, shape: jnp.ones(shape, dtype=x.dtype),
                (self.zmax + 2, d),
            )[species]
            if self.use_bias:
                b = self.param(
                    f"b_{i+1}",
                    lambda key, shape: jnp.zeros(shape, dtype=x.dtype),
                    (self.zmax + 2, d),
                )[species]
            else:
                b = 0
            x = activation_from_str(self.activation)(sig * x + b)
        x = nn.Dense(
            self.neurons[-1],
            use_bias=self.use_bias,
            name=f"Layer_{len(self.neurons)}",
            kernel_init=kernel_init,
        )(x)
        sig = self.param(
            f"sig_{len(self.neurons)}",
            lambda key, shape: jnp.ones(shape, dtype=x.dtype),
            (self.zmax + 2, self.neurons[-1]),
        )[species]
        if self.use_bias:
            b = self.param(
                f"b_{len(self.neurons)}",
                lambda key, shape: jnp.zeros(shape, dtype=x.dtype),
                (self.zmax + 2, self.neurons[-1]),
            )[species]
        else:
            b = 0
        x = sig * x + b
        if self.squeeze and x.shape[-1] == 1:
            x = jnp.squeeze(x, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: x} if output_key is not None else x
        return x


class ZLoRANet(nn.Module):
    """A fully connected neural network module with Z-dependent low-rank adaptation.
    
    FID: ZLORANET
    """

    neurons: Sequence[int]
    """A sequence of integers representing the dimensions of the network."""
    ranks: Sequence[int]
    """A sequence of integers representing the ranks of the low-rank adaptation at each layer."""
    zmax: int = 86
    """The maximum atomic number to consider."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use."""
    use_bias: bool = True
    """Whether to use bias in the dense layers."""
    input_key: Optional[str] = None
    """The key of the input tensor."""
    output_key: Optional[str] = None
    """The key of the output tensor."""
    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method to use."""
    species_key: str = "species"
    """The key of the species tensor."""

    FID: ClassVar[str] = "ZLORANET"

    @nn.compact
    def __call__(self, inputs: Union[dict, jax.Array]) -> Union[dict, jax.Array]:
        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            species, x = inputs
        else:
            species, x = inputs[self.species_key], inputs[self.input_key]

        # activation = (
        #     activation_from_str(self.activation)
        #     if isinstance(self.activation, str)
        #     else self.activation
        # )
        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )
        ############################
        for i, d in enumerate(self.neurons[:-1]):
            xi = nn.Dense(
                d, use_bias=self.use_bias, name=f"Layer_{i+1}", kernel_init=kernel_init
            )(x)
            A = self.param(
                f"A_{i+1}",
                lambda key, shape: jnp.zeros(shape, dtype=x.dtype),
                (self.zmax + 2, self.ranks[i], x.shape[-1]),
            )[species]
            B = self.param(
                f"B_{i+1}",
                lambda key, shape: jnp.zeros(shape, dtype=x.dtype),
                (self.zmax + 2, d, self.ranks[i]),
            )[species]
            Ax = jnp.einsum("zrd,zd->zr", A, x)
            BAx = jnp.einsum("zrd,zd->zr", B, Ax)
            x = activation_from_str(self.activation)(xi + BAx)
        xi = nn.Dense(
            self.neurons[-1],
            use_bias=self.use_bias,
            name=f"Layer_{len(self.neurons)}",
            kernel_init=kernel_init,
        )(x)
        A = self.param(
            f"A_{len(self.neurons)}",
            lambda key, shape: jnp.zeros(shape, dtype=x.dtype),
            (self.zmax + 2, self.ranks[-1], x.shape[-1]),
        )[species]
        B = self.param(
            f"B_{len(self.neurons)}",
            lambda key, shape: jnp.zeros(shape, dtype=x.dtype),
            (self.zmax + 2, self.neurons[-1], self.ranks[-1]),
        )[species]
        Ax = jnp.einsum("zrd,zd->zr", A, x)
        BAx = jnp.einsum("zrd,zd->zr", B, Ax)
        x = xi + BAx
        if self.squeeze and x.shape[-1] == 1:
            x = jnp.squeeze(x, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: x} if output_key is not None else x
        return x


class BlockIndexNet(nn.Module):
    """Chemical-species-specific neural network using precomputed species index.

    FID: BLOCK_INDEX_NET

    A neural network that applies a species-specific fully connected network to each atom embedding.
    A species index must be provided to filter the embeddings for each species and apply the corresponding network.
    This index can be obtained using the SPECIES_INDEXER preprocessing module from `fennol.models.preprocessing.SpeciesIndexer`

    """

    output_dim: int
    """The dimension of the output of the fully connected networks."""
    hidden_neurons: Sequence[int]
    """The hidden dimensions of the fully connected networks.
        If a dictionary is provided, it should map species names to dimensions.
        If a sequence is provided, the same dimensions will be used for all species."""
    used_blocks: Optional[Sequence[str]] = None
    """The blocks to use. If None, all blocks will be used."""
    activation: Union[Callable, str] = "silu"
    """The activation function to use in the fully connected networks."""
    use_bias: bool = True
    """Whether to include bias terms in the fully connected networks."""
    input_key: Optional[str] = None
    """The key in the input dictionary that corresponds to the embeddings of the atoms."""
    block_index_key: str = "block_index"
    """The key in the input dictionary that corresponds to the block index of the atoms. See `fennol.models.preprocessing.BlockIndexer`"""
    output_key: Optional[str] = None
    """The key in the output dictionary that corresponds to the network's output."""

    squeeze: bool = False
    """Whether to remove the last axis of the output tensor if it is of dimension 1."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initialization method for the fully connected networks."""
    # check_unhandled: bool = True

    FID: ClassVar[str] = "BLOCK_INDEX_NET"

    # def setup(self):
    #     all_blocks = CHEMICAL_BLOCKS_NAMES
    #     if self.used_blocks is None:
    #         used_blocks = all_blocks
    #     else:
    #         used_blocks = []
    #         for b in self.used_blocks:
    #             b_=str(b).strip().upper()
    #             if b_ not in all_blocks:
    #                 raise ValueError(f"Block {b} not found in {all_blocks}")
    #             used_blocks.append(b_)
    #     used_blocks = set(used_blocks)
    #     self._used_blocks = used_blocks

    #     if not (
    #         isinstance(self.hidden_neurons, dict)
    #         or isinstance(self.hidden_neurons, FrozenDict)
    #     ):
    #         neurons = {k: self.hidden_neurons for k in used_blocks}
    #     else:
    #         neurons = {}
    #         for b in self.hidden_neurons.keys():
    #             b_=str(b).strip().upper()
    #             if b_ not in all_blocks:
    #                 raise ValueError(f"Block {b} does not exist.  Available blocks are {all_blocks}")
    #             neurons[b_] = self.hidden_neurons[b]
    #         used_blocks = set(neurons.keys())
    #         if used_blocks != self._used_blocks and self.used_blocks is not None:
    #             print(
    #                 f"Warning: hidden neurons definitions do not match specified used_blocks {self.used_blocks}. Using blocks defined in hidden_neurons.")
    #         self._used_blocks = used_blocks

    #     self.networks = {
    #         k: FullyConnectedNet(
    #             [*neurons[k], self.output_dim],
    #             self.activation,
    #             self.use_bias,
    #             name=k,
    #             kernel_init=self.kernel_init,
    #         )
    #         for k in self._used_blocks
    #     }

    @nn.compact
    def __call__(
        self, inputs: Union[dict, Tuple[jax.Array, jax.Array]]
    ) -> Union[dict, jax.Array]:

        if self.input_key is None:
            assert not isinstance(
                inputs, dict
            ), "input key must be provided if inputs is a dictionary"
            species, embedding, block_index = inputs
        else:
            species, embedding = inputs["species"], inputs[self.input_key]
            block_index = inputs[self.block_index_key]

        assert isinstance(
            block_index, dict
        ), "block_index must be a dictionary for BlockIndexNet"

        networks = {
            k: FullyConnectedNet(
                [*self.hidden_neurons, self.output_dim],
                self.activation,
                self.use_bias,
                name=k,
                kernel_init=self.kernel_init,
            )
            for k in block_index.keys()
        }

        ############################
        # initialization => instantiate all networks
        if self.is_initializing():
            x = jnp.zeros((1, embedding.shape[-1]), dtype=embedding.dtype)
            [net(x) for net in networks.values()]

        # if self.check_unhandled:
        #     for b in block_index.keys():
        #         if b not in networks.keys():
        #             raise ValueError(f"Block {b} not found in networks. Available blocks are {self.networks.keys()}")

        ############################
        outputs = []
        indices = []
        for s, net in networks.items():
            if s not in block_index:
                continue
            if block_index[s] is None:
                continue
            idx = block_index[s]
            o = net(embedding[idx])
            outputs.append(o)
            indices.append(idx)

        o = jnp.concatenate(outputs, axis=0)
        idx = jnp.concatenate(indices, axis=0)

        out = (
            jnp.zeros((species.shape[0], *o.shape[1:]), dtype=o.dtype)
            .at[idx]
            .set(o, mode="drop")
        )

        if self.squeeze and out.shape[-1] == 1:
            out = jnp.squeeze(out, axis=-1)
        ############################

        if self.input_key is not None:
            output_key = self.name if self.output_key is None else self.output_key
            return {**inputs, output_key: out} if output_key is not None else out
        return out


import jax
import jax.numpy as jnp
import flax.linen as nn
from ...utils.spherical_harmonics import generate_spherical_harmonics, CG_SO3
from ..misc.encodings import SpeciesEncoding, RadialBasis
import dataclasses
import numpy as np
from typing import Dict, ClassVar


class GaussianMomentsEmbedding(nn.Module):
    """Gaussian moments embedding

    The construction of this embedding is similar to ACE but with a fixed lmax=3 and
    a subset of tensor product paths chosen by hand.

    ### Reference 
    adapted from J. Chem. Theory Comput. 2020, 16, 8, 5410–5421
    (https://pubs.acs.org/doi/full/10.1021/acs.jctc.0c00347)

    """

    _graphs_properties: Dict
    nchannels: int = 7
    """The number of chemical-radial (chemrad) channels for the density representation."""
    graph_key: str = "graph"
    """The key in the input dictionary that corresponds to the molecular graph."""
    embedding_key: str = "embedding"
    """The key in the output dictionary where the computed embedding will be stored."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """A dictionary of parameters for the species encoding. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """A dictionary of parameters for the radial basis. See `fennol.models.misc.encodings.RadialBasis`"""

    FID: ClassVar[str] = "GAUSSIAN_MOMENTS"


    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert len(species.shape) == 1, "Species must be a 1D array (batches must be flattened)"

        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]

        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        radial_basis = RadialBasis(
            **{**self.radial_basis, "end": cutoff, "name": "RadialBasis"}
        )(graph["distances"])
        radial_size = radial_basis.shape[-1]

        species_encoding = SpeciesEncoding(**self.species_encoding,name="SpeciesEncoding")(species)
        afvs_size = species_encoding.shape[-1]

        chemrad_coupling = self.param(
            "chemrad_coupling",
            nn.initializers.normal(stddev=1.0 / (afvs_size * radial_size) ** 0.5),
            (afvs_size, radial_size, self.nchannels),
        )
        xij = (
            jnp.einsum(
                "ai,aj,ijk->ak",
                species_encoding[edge_dst],
                radial_basis,
                chemrad_coupling,
            )
            * graph["switch"][:, None]
        )

        Yij = generate_spherical_harmonics(lmax=3, normalize=False)(
            graph["vec"] / graph["distances"][:, None]
        )
        rhoij = xij[:, :, None] * Yij[:, None, :]

        rhoi = jax.ops.segment_sum(rhoij, edge_src, species.shape[0])

        xi0 = jax.lax.index_in_dim(rhoi, 0, axis=-1, keepdims=False)

        rhoi1 = jax.lax.dynamic_slice_in_dim(rhoi, start_index=1, slice_size=3, axis=-1)
        rhoi2 = jax.lax.dynamic_slice_in_dim(rhoi, start_index=4, slice_size=5, axis=-1)
        rhoi3 = jax.lax.dynamic_slice_in_dim(rhoi, start_index=9, slice_size=7, axis=-1)

        pairs = []
        triplets = []
        for i in range(self.nchannels):
            for j in range(i, self.nchannels):
                pairs.append([i, j])
                for k in range(j, self.nchannels):
                    triplets.append([i, j, k])

        p1, p2 = np.array(pairs).T
        p1, p2 = jnp.array(p1), jnp.array(p2)
        xi11 = jnp.sum(rhoi1[:, p1, :] * rhoi1[:, p2, :], axis=-1)
        xi22 = jnp.sum(rhoi2[:, p1, :] * rhoi2[:, p2, :], axis=-1)
        xi33 = jnp.sum(rhoi3[:, p1, :] * rhoi3[:, p2, :], axis=-1)

        t1, t2, t3 = np.array(triplets).T
        t1, t2, t3 = jnp.array(t1), jnp.array(t2), jnp.array(t3)
        rhoi2t1 = rhoi2[:, t1, :]
        rhoi1t2 = rhoi1[:, t2, :]
        rhoi1t3 = rhoi1[:, t3, :]
        w112 = jnp.array(CG_SO3(1, 1, 2))
        xi211 = jnp.einsum("...m,...n,...o,nom->...", rhoi2t1, rhoi1t2, rhoi1t3, w112)

        rhoi2t2 = rhoi2[:, t2, :]
        rhoi2t3 = rhoi2[:, t3, :]
        w222 = jnp.array(CG_SO3(2, 2, 2))
        xi222 = jnp.einsum("...m,...n,...o,nom->...", rhoi2t1, rhoi2t2, rhoi2t3, w222)

        rhoi3t1 = rhoi3[:, t1, :]
        w123 = jnp.array(CG_SO3(1, 2, 3))
        xi312 = jnp.einsum("...m,...n,...o,nom->...", rhoi3t1, rhoi1t2, rhoi2t3, w123)

        rhoi3t3 = rhoi3[:, t3, :]
        w233 = jnp.array(CG_SO3(2, 3, 3))
        xi323 = jnp.einsum("...m,...n,...o,nom->...", rhoi3t1, rhoi2t2, rhoi3t3, w233)

        embedding = jnp.concatenate(
            [species_encoding, xi0, xi11, xi22, xi33, xi211, xi222, xi312, xi323],
            axis=-1,
        )

        if self.embedding_key is None:
            return embedding
        return {**inputs, self.embedding_key: embedding}



import jax
import jax.numpy as jnp
import flax.linen as nn
import dataclasses
import numpy as np
from typing import Any, Dict, Union, Callable, Sequence, Optional, ClassVar

from ...utils.initializers import initializer_from_str
from ..misc.nets import GatedPerceptron


class CHGNetEmbedding(nn.Module):
    """ Crystal Hamiltonian Graph Neural Network

    FID: CHGNET

    ### Reference
    Deng, B., Zhong, P., Jun, K. et al. CHGNet as a pretrained universal neural network potential for charge-informed atomistic modelling.
        Nat Mach Intell 5, 1031–1041 (2023). https://doi.org/10.1038/s42256-023-00716-3
    
    """
    _graphs_properties: Dict
    dim: int = 64
    """The dimension of the embedding."""
    nmax_angle: int = 4
    """ The maximum fourier order for the angle representation."""
    nlayers: int = 3
    """The number of layers."""
    graph_key: str = "graph"
    """The key for the graph input."""
    graph_angle_key: Optional[str] = None
    """The key for the angular graph input."""
    embedding_key: str = "embedding"
    """The key for the embedding output."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`"""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """The radial basis parameters. See `fennol.models.misc.encodings.RadialBasis`"""
    radial_basis_angle: Optional[dict] = None
    """The radial basis parameters for angle embedding. See `fennol.models.misc.encodings.RadialBasis`"""
    keep_all_layers: bool = False
    """Whether to keep all layers in the output."""
    kernel_init: Union[str, Callable] = "lecun_normal()"
    """The kernel initializer for Dense operations."""

    FID: ClassVar[str]  = "CHGNET"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        assert (
            len(species.shape) == 1
        ), "Species must be a 1D array (batches must be flattened)"

        kernel_init = (
            initializer_from_str(self.kernel_init)
            if isinstance(self.kernel_init, str)
            else self.kernel_init
        )

        ##################################################
        # Check that the graph_angle is a subgraph of graph
        graph = inputs[self.graph_key]
        graph_angle_key = (
            self.graph_angle_key if self.graph_angle_key is not None else self.graph_key
        )
        graph_angle = inputs[graph_angle_key]

        correct_graph = (
            graph_angle_key == self.graph_key
            or self._graphs_properties[graph_angle_key]["parent_graph"]
            == self.graph_key
        )
        assert (
            correct_graph
        ), f"graph_angle_key={graph_angle_key} must be a subgraph of graph_key={self.graph_key}"
        assert "angles" in graph_angle, f"Graph {graph_angle_key} must contain angles"
        # check if graph_angle is a filtered graph
        filtered = "parent_graph" in self._graphs_properties[graph_angle_key]
        if filtered:
            filter_indices = graph_angle["filter_indices"]

        ##################################################
        ### SPECIES ENCODING ###
        zi = SpeciesEncoding(**self.species_encoding, name="SpeciesEncoding")(species)

        vi = nn.Dense(self.dim, name="vi0", use_bias=True, kernel_init=kernel_init)(zi)

        ##################################################
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        distances = graph["distances"]
        switch = graph["switch"][:, None]

        ### COMPUTE RADIAL BASIS ###
        radial_basis = RadialBasis(
            **{
                **self.radial_basis,
                "end": cutoff,
                "name": f"RadialBasis",
            }
        )(distances)

        ##################################################
        ### GET ANGLES ###
        angles = graph_angle["angles"][:, None]
        angle_src, angle_dst = graph_angle["angle_src"], graph_angle["angle_dst"]
        switch_angles = graph_angle["switch"][:, None]
        central_atom = graph_angle["central_atom"]

        ### COMPUTE RADIAL BASIS FOR ANGLES ###
        if self.radial_basis_angle is not None:
            dangles = graph_angle["distances"]
            radial_basis_angle = (
                RadialBasis(
                    **{
                        **self.radial_basis_angle,
                        "end": self._graphs_properties[graph_angle_key]["cutoff"],
                        "name": f"RadialBasisAngle",
                    }
                )(dangles)
                * switch_angles
            )

        else:
            if filtered:
                radial_basis_angle = radial_basis[filter_indices] * switch_angles
            else:
                radial_basis_angle = radial_basis * switch

        radial_basis = radial_basis * switch

        eij, eija = jnp.split(
            nn.Dense(
                2 * self.dim, use_bias=False, kernel_init=kernel_init, name="eij0"
            )(radial_basis),
            2,
            axis=-1,
        )

        eijb = nn.Dense(self.dim, use_bias=False, kernel_init=kernel_init, name="eijb")(
            radial_basis_angle
        )
        eijkb = eijb[angle_src] * eijb[angle_dst]

        ##################################################
        ### ANGULAR BASIS ###
        # build fourier series for angles
        nangles = jnp.asarray(
            np.arange(self.nmax_angle + 1, dtype=distances.dtype)[None, :]
        )

        Ac = jnp.cos(nangles * angles)
        As = jnp.sin(nangles[:, 1:] * angles)
        aijk = nn.Dense(
            self.dim, use_bias=False, kernel_init=kernel_init, name="aijk0"
        )(jnp.concatenate([Ac, As], axis=-1))
        aikj = aijk

        ##################################################
        if self.keep_all_layers:
            vis = []

        ### LOOP OVER LAYERS ###
        for layer in range(self.nlayers):
            phiv = GatedPerceptron(
                self.dim,
                use_bias=True,
                kernel_init=kernel_init,
                activation=nn.silu,
                name=f"phiv{layer+1}",
            )(jnp.concatenate([vi[edge_src], vi[edge_dst], eij], axis=-1))

            vi = vi + nn.Dense(
                self.dim, use_bias=True, kernel_init=kernel_init, name=f"vi{layer+1}"
            )(jax.ops.segment_sum(phiv * eija, edge_src, vi.shape[0]))

            if self.keep_all_layers:
                vis.append(vi)
                
            if layer == self.nlayers - 1:
                break
                
            eij_ang = eij[filter_indices] if filtered else eij
            eij_angsrc = eij_ang[angle_src]
            eij_angdst = eij_ang[angle_dst]

            phie = GatedPerceptron(
                self.dim,
                use_bias=True,
                kernel_init=kernel_init,
                activation=nn.silu,
                name=f"phie{layer+1}",
            )

            vi_ang = vi[central_atom]
            phie_ijk = phie(
                jnp.concatenate(
                    [eij_angsrc, eij_angdst, aijk, vi_ang],
                    axis=-1,
                )
            )
            phie_ikj = phie(
                jnp.concatenate(
                    [eij_angdst, eij_angsrc, aikj, vi_ang],
                    axis=-1,
                )
            )

            eij_ang = eij_ang + nn.Dense(
                self.dim, use_bias=False, kernel_init=kernel_init, name=f"Le{layer+1}"
            )(
                jax.ops.segment_sum(phie_ikj * eijkb, angle_dst, eij_ang.shape[0])
                + jax.ops.segment_sum(phie_ijk * eijkb, angle_src, eij_ang.shape[0])
            )
            eij_angsrc = eij_ang[angle_src]
            eij_angdst = eij_ang[angle_dst]
            eij = eij.at[filter_indices].set(eij_ang)

            phia = GatedPerceptron(
                self.dim,
                use_bias=True,
                kernel_init=kernel_init,
                activation=nn.silu,
                name=f"phia{layer+1}",
            )
            aijk = aijk + phia(
                jnp.concatenate([eij_angsrc, eij_angdst, aijk, vi_ang], axis=-1)
            )
            aikj = aikj + phia(
                jnp.concatenate([eij_angdst, eij_angsrc, aikj, vi_ang], axis=-1)
            )

            

        output = {
            **inputs,
            self.embedding_key: vi,
        }
        if self.keep_all_layers:
            output[self.embedding_key + "_layers"] = jnp.stack(vis, axis=1)
        return output



import jax
import jax.numpy as jnp
import flax.linen as nn
from typing import Sequence, Dict, Callable, ClassVar,Union
import numpy as np
from ...utils.periodic_table import PERIODIC_TABLE
from ..misc.encodings import SpeciesEncoding
from ..misc.nets import FullyConnectedNet
import dataclasses


class AIMNet(nn.Module):
    """Atom-In-Molecule Network message-passing embedding

    FID : AIMNET

    ### Reference
    Roman Zubatyuk et al. ,Accurate and transferable multitask prediction of chemical properties with an atoms-in-molecules neural network.Sci. Adv.5,eaav6490(2019).DOI:10.1126/sciadv.aav6490
    """

    _graphs_properties: Dict
    graph_angle_key: str
    """ The key in the input dictionary that corresponds to the angular graph. """
    nlayers: int = 3
    """ The number of message-passing layers."""
    zmax: int = 86
    """ The maximum atomic number to allocate AFV."""
    radial_eta: float = 16.0
    """ Controls the width of the gaussian sensity functions in radial AEV."""
    angular_eta: float = 8.0
    """ Controls the width of the gaussian sensity functions in angular AEV."""
    radial_dist_divisions: int = 16
    """ Number of basis function to encode ditance in radial AEV."""
    angular_dist_divisions: int = 4
    """ Number of basis function to encode ditance in angular AEV."""
    zeta: float = 32.0
    """ The power parameter in angle embedding."""
    angle_sections: int = 4
    """ The number of angle sections."""
    radial_start: float = 0.8
    """ The starting distance in radial AEV."""
    angular_start: float = 0.8
    """ The starting distance in angular AEV."""
    embedding_key: str = "embedding"
    """ The key to use for the output embedding in the returned dictionary."""
    graph_key: str = "graph"
    """ The key in the input dictionary that corresponds to the radial graph."""
    keep_all_layers: bool = False
    """ If True, the output will contain the embeddings from all layers."""

    activation: Union[Callable, str] = "swish"
    """ The activation function to use."""
    combination_neurons: Sequence[int] = dataclasses.field(
        default_factory=lambda: [256, 128, 16]
    )
    """ The number of neurons in the AFV combination network."""

    embedding_neurons: Sequence[int] = dataclasses.field(
        default_factory=lambda: [512, 256, 256]
    )
    """ The number of neurons in the embedding network."""
    interaction_neurons: Sequence[int] = dataclasses.field(
        default_factory=lambda: [256, 256, 128]
    )
    """ The number of neurons in the interaction network."""
    afv_neurons: Sequence[int] = dataclasses.field(
        default_factory=lambda: [256, 256, 16]
    )
    """ The number of neurons in the AFV update network. The last number of neurons defines the size of AFV."""

    FID: ClassVar[str] = "AIMNET"

    @nn.compact
    def __call__(self, inputs):
        """Forward pass of the AIMNet model."""
        species = inputs["species"]

        # species encoding (AFV)
        afv_dim = self.afv_neurons[-1]
        afv = SpeciesEncoding(dim=afv_dim, zmax=self.zmax, encoding="random")(species)

        # Radial graph
        graph = inputs[self.graph_key]
        distances = graph["distances"]
        switch = graph["switch"]
        edge_src = graph["edge_src"]
        edge_dst = graph["edge_dst"]

        # Radial AEV
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        shiftR = jnp.asarray(
            np.linspace(self.radial_start, cutoff, self.radial_dist_divisions + 1)[
                None, :-1
            ],
            dtype=distances.dtype,
        )
        x2 = self.radial_eta * (distances[:, None] - shiftR) ** 2
        radial_terms = 0.25 * jnp.exp(-x2) * switch[:, None]

        # Angular graph
        graph = inputs[self.graph_angle_key]
        edge_dst_a = graph["edge_dst"]
        angles = graph["angles"]
        distances = graph["distances"]
        central_atom = graph["central_atom"]
        angle_src, angle_dst = graph["angle_src"], graph["angle_dst"]
        angle_atom_1 = edge_dst_a[angle_src]
        angle_atom_2 = edge_dst_a[angle_dst]
        switch = graph["switch"]
        d12 = 0.5 * (distances[angle_src] + distances[angle_dst])[:, None]

        # Angular AEV parameters
        angular_cutoff = self._graphs_properties[self.graph_angle_key]["cutoff"]
        angle_start = np.pi / (2 * self.angle_sections)
        shiftZ = jnp.asarray(
            (np.linspace(0, np.pi, self.angle_sections + 1) + angle_start)[None, :-1],
            dtype=distances.dtype,
        )
        shiftA = jnp.asarray(
            np.linspace(
                self.angular_start, angular_cutoff, self.angular_dist_divisions + 1
            )[None, :-1],
            dtype=distances.dtype,
        )

        # Angular AEV
        factor1 = (0.5 + 0.5 * jnp.cos(angles[:, None] - shiftZ)) ** self.zeta
        factor2 = jnp.exp(-self.angular_eta * (d12 - shiftA) ** 2)
        angular_terms = (
            (factor1[:, None, :] * factor2[:, :, None]).reshape(
                -1, self.angle_sections * self.angular_dist_divisions
            )
            * 2
            * (switch[angle_src] * switch[angle_dst])[:, None]
        )

        if self.keep_all_layers:
            mis = []
        for layer in range(self.nlayers):
            # combine pair info
            Gij = (radial_terms[:, None, :] * afv[edge_dst, :, None]).reshape(
                radial_terms.shape[0], -1
            )
            Gri = jax.ops.segment_sum(Gij, edge_src, species.shape[0])

            # combine triplet info
            afv1 = afv[angle_atom_1]
            afv2 = afv[angle_atom_2]
            afv12 = jnp.concatenate((afv1 * afv2, afv1 + afv2), axis=-1)
            afv_ang = FullyConnectedNet(
                self.combination_neurons,
                activation=self.activation,
                name=f"combination_net_{layer}",
            )(afv12)
            Gijk = (angular_terms[:, None, :] * afv_ang[:, :, None]).reshape(
                angular_terms.shape[0], -1
            )
            Gai = jax.ops.segment_sum(Gijk, central_atom, species.shape[0])

            # environment field
            fi = FullyConnectedNet(
                self.embedding_neurons,
                activation=self.activation,
                name=f"embedding_net_{layer}",
            )(jnp.concatenate((Gri, Gai), axis=-1))

            # update AFV
            dafv = FullyConnectedNet(
                self.afv_neurons,
                activation=self.activation,
                name=f"afv_update_net_{layer}",
            )(fi)
            afv = afv + dafv

            if self.keep_all_layers or layer == self.nlayers - 1:
                # embedding
                mi = FullyConnectedNet(
                    self.interaction_neurons,
                    activation=self.activation,
                    name=f"interaction_net_{layer}",
                )(fi)
                if self.keep_all_layers:
                    mis.append(mi)

        embedding_key = (
            self.embedding_key if self.embedding_key is not None else self.name
        )

        output = {**inputs, embedding_key: mi, embedding_key + "_afv": afv}
        if self.keep_all_layers:
            output[embedding_key + "_layers"] = jnp.stack(mis, axis=1)

        return output



import jax
import jax.numpy as jnp
import flax.linen as nn
import numpy as np

# from jaxopt.linear_solve import solve_cg, solve_iterative_refinement, solve_gmres
from typing import Any, Dict, Union, Callable, Sequence, Optional, ClassVar
from ...utils import AtomicUnits as au
import dataclasses
from ...utils.periodic_table import (
    D3_ELECTRONEGATIVITIES,
    D3_HARDNESSES,
    D3_VDW_RADII,
    D3_COV_RADII,
    D3_KAPPA,
    VDW_RADII,
    POLARIZABILITIES,
    VALENCE_ELECTRONS,
)
import math


def prepare_reciprocal_space(
    cells, reciprocal_cells, coordinates, batch_index, k_points, bewald
):
    """Prepare variables for Ewald summation in reciprocal space"""
    A = reciprocal_cells
    if A.shape[0] == 1:
        s = coordinates @ A[0]
        ks = 2j * jnp.pi * jnp.einsum("ai,ki-> ak", s, k_points[0])  # nat x nk
    else:
        s = jnp.einsum("aj,aji->ai", coordinates,A[batch_index])
        ks = (
            2j * jnp.pi * jnp.einsum("ai,aki-> ak", s, k_points[batch_index])
        )  # nat x nk

    m2 = jnp.sum(
        jnp.einsum("ski,sji->skj", k_points, A) ** 2,
        axis=-1,
    )  # nsys x nk
    a2 = (jnp.pi / bewald) ** 2
    expfac = jnp.exp(-a2 * m2) / m2  # nsys x nk

    volume = jnp.abs(jnp.linalg.det(cells))  # nsys
    phiscale = (au.BOHR / jnp.pi) / volume
    selfscale = bewald * (2 * au.BOHR / jnp.pi**0.5)
    return batch_index, k_points, phiscale, selfscale, expfac, ks


def ewald_reciprocal(q, batch_index, k_points, phiscale, selfscale, expfac, ks):
    """Compute Coulomb interactions in reciprocal space using Ewald summation"""
    if phiscale.shape[0] == 1:
        Sm = (q[:, None] * jnp.exp(ks)).sum(axis=0)[None, :]  # nys x nk
    else:
        Sm = jax.ops.segment_sum(
            q[:, None] * jnp.exp(ks), batch_index, k_points.shape[0]
        )  # nsys x nk

    ### compute reciprocal Coulomb potential (https://arxiv.org/abs/1805.10363)
    phi = (
        jnp.real(((Sm * expfac)[batch_index] * jnp.exp(-ks)).sum(axis=-1))
        * phiscale[batch_index]
        - q * selfscale
    )

    return 0.5 * q * phi, phi


# def ewald_reciprocal(
#     q, cells, reciprocal_cells, coordinates, batch_index, k_points, bewald
# ):
#     A = reciprocal_cells
#     ### Ewald reciprocal space

#     if A.shape[0] == 1:
#         s = jnp.einsum("ij,aj->ai", A[0], coordinates)
#         ks = 2j * jnp.pi * jnp.einsum("ai,ki-> ak", s, k_points[0])  # nat x nk
#         Sm = (q[:, None] * jnp.exp(ks)).sum(axis=0)[None, :]  # nys x nk
#     else:
#         s = jnp.einsum("aij,aj->ai", A[batch_index], coordinates)
#         ks = (
#             2j * jnp.pi * jnp.einsum("ai,aki-> ak", s, k_points[batch_index])
#         )  # nat x nk
#         Sm = jax.ops.segment_sum(
#             q[:, None] * jnp.exp(ks), batch_index, k_points.shape[0]
#         )  # nsys x nk

#     m2 = jnp.sum(
#         jnp.einsum("sij,ski->skj", A, k_points) ** 2,
#         axis=-1,
#     )  # nsys x nk
#     a2 = (jnp.pi / bewald) ** 2
#     expfac = Sm * jnp.exp(-a2 * m2) / m2  # nsys x nk
#     volume = jnp.linalg.det(cells)  # nsys

#     ### compute reciprocal Coulomb potential (https://arxiv.org/abs/1805.10363)
#     phi = jnp.real((expfac[batch_index] * jnp.exp(-ks)).sum(axis=-1)) * (
#         (au.BOHR / jnp.pi) / volume[batch_index]
#     ) - q * (bewald * (2 * au.BOHR / jnp.pi**0.5))

#     return 0.5 * q * phi, phi


class Coulomb(nn.Module):
    """Coulomb interaction between distributed point charges

    FID: COULOMB   
    
    """
    _graphs_properties: Dict
    graph_key: str = "graph"
    """Key for the graph in the inputs"""
    charges_key: str = "charges"
    """Key for the charges in the inputs"""
    energy_key: Optional[str] = None
    """Key for the energy in the outputs"""
    # switch_fraction: float = 0.9
    scale: Optional[float] = None
    """Scaling factor for the energy"""
    charge_scale: Optional[float] = None
    """Scaling factor for the charges"""
    damp_style: Optional[str] = None
    """Damping style. Available options are: None, 'TS', 'OQDO', 'D3', 'SPOOKY', 'CP', 'KEY'"""
    damp_params: Dict = dataclasses.field(default_factory=dict)
    """Damping parameters"""
    bscreen: float = -1.0
    """Screening parameter. If >0, the Coulomb potential becomes a Yukawa potential and the reciprocal space is not computed"""
    trainable: bool = True
    """Whether the parameters are trainable"""
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""

    FID: ClassVar[str] = "COULOMB"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        distances = graph["distances"]
        switch = graph["switch"]

        rij = distances / au.BOHR
        q = inputs[self.charges_key]
        if q.shape[-1] == 1:
            q = jnp.squeeze(q, axis=-1)
        if self.charge_scale is not None:
            if self.trainable:
                charge_scale = jnp.abs(
                    self.param(
                        "charge_scale", lambda key: jnp.asarray(self.charge_scale)
                    )
                )
            else:
                charge_scale = self.charge_scale
            q = q * charge_scale

        damp_style = self.damp_style.upper() if self.damp_style is not None else None

        do_recip = self.bscreen <= 0.0 and "k_points" in graph

        if self.bscreen > 0.0:
            # dirfact = jax.scipy.special.erfc(self.bscreen * distances)
            dirfact = jnp.exp(-self.bscreen * distances)
        elif do_recip:
            k_points = graph["k_points"]
            bewald = graph["b_ewald"]
            cells = inputs["cells"]
            reciprocal_cells = inputs["reciprocal_cells"]
            batch_index = inputs["batch_index"]
            erec, _ = ewald_reciprocal(
                q,
                *prepare_reciprocal_space(
                    cells,
                    reciprocal_cells,
                    inputs["coordinates"],
                    batch_index,
                    k_points,
                    bewald,
                ),
            )
            dirfact = jax.scipy.special.erfc(bewald * distances)
        else:
            dirfact = 1.0

        if damp_style is None:
            Aij = switch * dirfact / rij
            eat = (
                0.5
                * q
                * jax.ops.segment_sum(Aij * q[edge_dst], edge_src, species.shape[0])
            )

        elif damp_style == "TS":
            cpB = self.damp_params.get("cpB", 3.5)
            s = self.damp_params.get("s", 2.4)

            if self.trainable:
                cpB = jnp.abs(self.param("cpB", lambda key: jnp.asarray(cpB)))
                s = jnp.abs(self.param("s", lambda key: jnp.asarray(s)))

            ratiovol_key = self.damp_params.get("ratiovol_key", None)
            if ratiovol_key is not None:
                ratiovol = inputs[ratiovol_key]
                if ratiovol.shape[-1] == 1:
                    ratiovol = jnp.squeeze(ratiovol, axis=-1)
                rvdw = jnp.asarray(VDW_RADII)[species] * ratiovol ** (1.0 / 3.0)
            else:
                rvdw = jnp.asarray(VDW_RADII)[species]
            Rij = rvdw[edge_src] + rvdw[edge_dst]
            Bij = cpB * (rij / Rij) ** s

            eBij = jnp.where(Bij < 20.0, jnp.exp(-Bij), 0.0)

            Aij = (dirfact - eBij) / rij * switch
            eat = (
                0.5
                * q
                * jax.ops.segment_sum(Aij * q[edge_dst], edge_src, species.shape[0])
            )

        elif damp_style == "OQDO":
            ratiovol_key = self.damp_params.get("ratiovol_key", None)
            alpha = jnp.asarray(POLARIZABILITIES)[species]
            if ratiovol_key is not None:
                ratiovol = inputs[ratiovol_key]
                if ratiovol.shape[-1] == 1:
                    ratiovol = jnp.squeeze(ratiovol, axis=-1)
                alpha = alpha * ratiovol

            alphaij = 0.5 * (alpha[edge_src] + alpha[edge_dst])
            Re = (alphaij * (128.0 / au.FSC ** (4.0 / 3.0))) ** (1.0 / 7.0)
            Re2 = Re**2
            Re4 = Re**4
            muw = (
                3.66316787e01
                - 5.79579187 * Re
                + 3.02674813e-01 * Re2
                - 3.65461255e-04 * Re4
            ) / (-1.46169102e01 + 7.32461225 * Re)
            # muw = (
            #     4.83053463e-01
            #     - 3.76191669e-02 * Re
            #     + 1.27066988e-03 * Re2
            #     - 7.21940151e-07 * Re4
            # ) / (3.84212120e-02 - 3.16915319e-02 * Re + 2.37410890e-02 * Re2)
            Bij = 0.5 * muw * rij**2

            eBij = jnp.where(Bij < 20.0, jnp.exp(-Bij), 0.0)

            Aij = (dirfact - eBij) / rij * switch
            eat = (
                0.5
                * q
                * jax.ops.segment_sum(Aij * q[edge_dst], edge_src, species.shape[0])
            )

        elif damp_style == "D3":
            ratiovol_key = self.damp_params.get("ratiovol_key", None)
            if ratiovol_key is not None:
                ratiovol = inputs[ratiovol_key]
                if ratiovol.shape[-1] == 1:
                    ratiovol = jnp.squeeze(ratiovol, axis=-1)
            else:
                ratiovol = 1.0

            gamma_scheme = self.damp_params.get("gamma_scheme", "D3")
            if gamma_scheme == "D3":
                if self.trainable:
                    rvdw = jnp.abs(
                        self.param("rvdw", lambda key: jnp.asarray(VDW_RADII))
                    )[species]
                else:
                    rvdw = jnp.asarray(VDW_RADII)[species]
                rvdw = rvdw * ratiovol ** (1.0 / 3.0)
                ai2 = rvdw**2
                gamma_ij = (ai2[edge_src] + ai2[edge_dst] + 1.0e-3) ** (-0.5)

            elif gamma_scheme == "QDO":
                gscale = self.damp_params.get("gamma_scale", 2.0)
                if self.trainable:
                    gscale = jnp.abs(
                        self.param("gamma_scale", lambda key: jnp.asarray(gscale))
                    )
                    alpha = jnp.abs(
                        self.param("alpha", lambda key: jnp.asarray(POLARIZABILITIES))
                    )[species]
                else:
                    alpha = jnp.asarray(POLARIZABILITIES)[species]
                alpha = alpha * ratiovol
                alphaij = 0.5 * (alpha[edge_src] + alpha[edge_dst])
                rvdwij = (alphaij * (128.0 / au.FSC ** (4.0 / 3.0))) ** (1.0 / 7.0)
                gamma_ij = gscale / rvdwij
            else:
                raise NotImplementedError(
                    f"gamma_scheme {gamma_scheme} not implemented"
                )

            Aij = (dirfact - jax.scipy.special.erfc(gamma_ij * rij)) / rij * switch

            eat = (
                0.5
                * q
                * jax.ops.segment_sum(Aij * q[edge_dst], edge_src, species.shape[0])
            )

        elif damp_style == "SPOOKY":
            shortrange_cutoff = self.damp_params.get("shortrange_cutoff", 5.0)
            r_on = self.damp_params.get("r_on", 0.25) * shortrange_cutoff
            r_off = self.damp_params.get("r_off", 0.75) * shortrange_cutoff
            x1 = (distances - r_on) / (r_off - r_on)
            x2 = 1.0 - x1
            mask1 = x1 <= 1.0e-6
            mask2 = x2 <= 1.0e-6
            x1 = jnp.where(mask1, 1.0, x1)
            x2 = jnp.where(mask2, 1.0, x2)
            s1 = jnp.where(mask1, 0.0, jnp.exp(-1.0 / x1))
            s2 = jnp.where(mask2, 0.0, jnp.exp(-1.0 / x2))
            Bij = s2 / (s1 + s2)

            Aij = Bij / (rij**2 + 1) ** 0.5 + (dirfact - Bij) / rij * switch
            eat = (
                0.5
                * q
                * jax.ops.segment_sum(Aij * q[edge_dst], edge_src, species.shape[0])
            )

        elif damp_style == "CP":
            cpA = self.damp_params.get("cpA", 4.42)
            cpB = self.damp_params.get("cpB", 4.12)
            gamma = self.damp_params.get("gamma", 0.5)

            if self.trainable:
                cpA = jnp.abs(self.param("cpA", lambda key: jnp.asarray(cpA)))
                cpB = jnp.abs(self.param("cpB", lambda key: jnp.asarray(cpB)))
                gamma = jnp.abs(self.param("gamma", lambda key: jnp.asarray(gamma)))

            rvdw = jnp.asarray(VDW_RADII)[species]
            ratiovol_key = self.damp_params.get("ratiovol_key", None)
            if ratiovol_key is not None:
                ratiovol = inputs[ratiovol_key]
                if ratiovol.shape[-1] == 1:
                    ratiovol = jnp.squeeze(ratiovol, axis=-1)
                rvdw = rvdw * ratiovol ** (1.0 / 3.0)

            Zv = jnp.asarray(VALENCE_ELECTRONS)[species]
            Zi, Zj = Zv[edge_src], Zv[edge_dst]
            qi, qj = q[edge_src], q[edge_dst]
            rvdwi, rvdwj = rvdw[edge_src], rvdw[edge_dst]

            eAi = jnp.exp(-cpA * rij / rvdwi)
            eAj = jnp.exp(-cpA * rij / rvdwj)
            eBi = jnp.exp(-cpB * rij / rvdwi)
            eBj = jnp.exp(-cpB * rij / rvdwj)
            eBij = eBi * eBj - eBi - eBj

            Bshort = jnp.exp(-gamma * distances**4)
            Dshort = 1.0 - Bshort
            ecp = Dshort * (
                Zi * Zj * (eAi + eAj + eBij)
                - qi * Zj * (eAi + eBij)
                - qj * Zi * (eAj + eBij)
            )

            # eq = qi * qj * (1 + eBij) * (1 - Bshort)
            eqq = qi * qj * (dirfact - Bshort + eBij * Dshort)

            epair = (ecp + eqq) * switch / rij

            # epair = (
            #     (1 - Bshort)
            #     * (
            #         Zi * Zj * (eAi + eAj + eBij)
            #         - qi * Zj * (eAi + eBij)
            #         - qj * Zi * (eAj + eBij)
            #         + qi * qj * (1 + eBij)
            #     )
            #     * switch
            #     / rij
            # )
            eat = 0.5 * jax.ops.segment_sum(epair, edge_src, species.shape[0])

        elif damp_style == "KEY":
            damp_key = self.damp_params["key"]
            damp = inputs[damp_key]
            epair = (dirfact - damp) * switch / rij
            eat = 0.5 * jax.ops.segment_sum(epair, edge_src, species.shape[0])
        else:
            raise NotImplementedError(f"damp_style {self.damp_style} not implemented")

        if do_recip:
            eat = eat + erec

        if self.scale is not None:
            if self.trainable:
                scale = jnp.abs(
                    self.param("scale", lambda key: jnp.asarray(self.scale))
                )
            else:
                scale = self.scale
            eat = eat * scale

        energy_key = self.energy_key if self.energy_key is not None else self.name
        energy_unit = au.get_multiplier(self._energy_unit)
        out = {**inputs, energy_key: eat*energy_unit}
        if do_recip:
            out[energy_key + "_reciprocal"] = erec*energy_unit
        return out


class QeqD4(nn.Module):
    """ QEq-D4 charge equilibration scheme

    FID: QEQ_D4
    
    ### Reference
    E. Caldeweyher et al.,A generally applicable atomic-charge dependent London dispersion correction,
    J Chem Phys. 2019 Apr 21;150(15):154122. (https://doi.org/10.1063/1.5090222)
    """
    graph_key: str = "graph"
    """Key for the graph in the inputs"""
    trainable: bool = False
    """Whether the parameters are trainable"""
    charges_key: str = "charges"
    """Key for the charges in the outputs. 
        If charges are provided in the inputs, they are not re-optimized and we only compute the energy"""
    energy_key: Optional[str] = None
    """Key for the energy in the outputs"""
    chi_key: Optional[str] = None
    """Key for additional electronegativity in the inputs"""
    c3_key: Optional[str] = None
    """Key for additional c3 in the inputs. Only used if charges are provided in the inputs"""
    c4_key: Optional[str] = None
    """Key for additional c4 in the inputs. Only used if charges are provided in the inputs"""
    total_charge_key: str = "total_charge"
    """Key for the total charge in the inputs"""
    non_interacting_guess: bool = False
    """Whether to use the non-interacting limit as an initial guess."""
    solver: str = "gmres"
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""

    FID: ClassVar[str] = "QEQ_D4"

    @nn.compact
    def __call__(self, inputs):
        species = inputs["species"]
        graph = inputs[self.graph_key]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        switch = graph["switch"]

        rij = graph["distances"] / au.BOHR

        do_recip = "k_points" in graph
        if do_recip:
            k_points = graph["k_points"]
            bewald = graph["b_ewald"]
            cells = inputs["cells"]
            reciprocal_cells = inputs["reciprocal_cells"]
            batch_index = inputs["batch_index"]
            dirfact = jax.scipy.special.erfc(bewald * graph["distances"])
            ewald_params = prepare_reciprocal_space(
                cells,
                reciprocal_cells,
                inputs["coordinates"],
                batch_index,
                k_points,
                bewald,
            )
        else:
            dirfact = 1.0

        Jii = D3_HARDNESSES
        ai = D3_VDW_RADII
        ETA = [jii + (2.0 / np.pi) ** 0.5 / aii for jii, aii in zip(Jii, ai)]

        # D3 parameters
        if self.trainable:
            ENi = self.param("EN", lambda key: jnp.asarray(D3_ELECTRONEGATIVITIES))[
                species
            ]
            # Jii = self.param("J", lambda key: jnp.asarray(D3_HARDNESSES))[species]
            eta = jnp.abs(self.param("eta", lambda key: jnp.asarray(ETA)))[species]
            ai = jnp.abs(self.param("a", lambda key: jnp.asarray(D3_VDW_RADII)))[
                species
            ]
            rci = jnp.abs(self.param("rc", lambda key: jnp.asarray(D3_COV_RADII)))[
                species
            ]
            c3 = self.param("c3", lambda key: jnp.zeros(len(D3_ELECTRONEGATIVITIES)))[
                species
            ]
            c4 = jnp.abs(
                self.param("c4", lambda key: jnp.zeros(len(D3_ELECTRONEGATIVITIES)))[
                    species
                ]
            )
            kappai = self.param("kappa", lambda key: jnp.asarray(D3_KAPPA))[species]
            k1 = jnp.abs(self.param("k1", lambda key: jnp.asarray(7.5)))
            training = "training" in inputs.get("flags", {})
            if training:
                regularization = (
                    (ENi - jnp.asarray(D3_ELECTRONEGATIVITIES)[species]) ** 2
                    + (eta - jnp.asarray(ETA)[species]) ** 2
                    + (ai - jnp.asarray(D3_VDW_RADII)[species]) ** 2
                    + (rci - jnp.asarray(D3_COV_RADII)[species]) ** 2
                    + (kappai - jnp.asarray(D3_KAPPA)[species]) ** 2
                    + (k1 - 7.5) ** 2
                )
        else:
            c3 = jnp.zeros_like(species, dtype=jnp.float32)
            c4 = jnp.zeros_like(species, dtype=jnp.float32)
            ENi = jnp.asarray(D3_ELECTRONEGATIVITIES)[species]
            # Jii = jnp.asarray(D3_HARDNESSES)[species]
            eta = jnp.asarray(ETA)[species]
            ai = jnp.asarray(D3_VDW_RADII)[species]
            rci = jnp.asarray(D3_COV_RADII)[species]
            kappai = jnp.asarray(D3_KAPPA)[species]
            k1 = 7.5

        ai2 = ai**2
        rcij = (
            rci.at[edge_src].get(mode="fill", fill_value=1.0)
            + rci.at[edge_dst].get(mode="fill", fill_value=1.0)
            + 1.0e-3
        )
        mCNij = 1.0 + jax.scipy.special.erf(-k1 * (rij / rcij - 1))
        mCNi = 0.5 * jax.ops.segment_sum(mCNij * switch, edge_src, species.shape[0])
        chi = ENi - kappai * (mCNi + 1.0e-3) ** 0.5
        if self.chi_key is not None:
            chi = chi + inputs[self.chi_key]

        gamma_ij = (
            ai2.at[edge_src].get(mode="fill", fill_value=1.0)
            + ai2.at[edge_dst].get(mode="fill", fill_value=1.0)
            + 1.0e-3
        ) ** (-0.5)

        Aii = eta  # Jii + ((2.0 / np.pi) ** 0.5) / ai
        # Aij = jax.scipy.special.erf(gamma_ij * rij) / rij * switch
        Aij = (dirfact - jax.scipy.special.erfc(gamma_ij * rij)) / rij * switch

        if self.charges_key in inputs:
            q = inputs[self.charges_key]
            q_ = q
        else:
            nsys = inputs["natoms"].shape[0]
            batch_index = inputs["batch_index"]

            def matvec(x):
                l, q = jnp.split(x, (nsys,))
                Aq_self = Aii * q
                qdest = q.at[edge_dst].get(mode="fill", fill_value=0.0)
                Aq_pair = jax.ops.segment_sum(Aij * qdest, edge_src, species.shape[0])
                Aq = (
                    Aq_self
                    + Aq_pair
                    + l.at[batch_index].get(mode="fill", fill_value=0.0)
                )
                if do_recip:
                    _, phirec = ewald_reciprocal(q, *ewald_params)
                    Aq = Aq + phirec
                Al = jax.ops.segment_sum(q, batch_index, nsys)
                return jnp.concatenate((Al, Aq))

            Qtot = (
                inputs[self.total_charge_key].astype(chi.dtype).reshape(nsys)
                if self.total_charge_key in inputs
                else jnp.zeros(nsys, dtype=chi.dtype)
            )
            b = jnp.concatenate([Qtot, -chi])

            if self.non_interacting_guess:
                # build initial guess
                si = 1./Aii
                q0 = -chi*si
                qtot = jax.ops.segment_sum(q0,batch_index,nsys)
                sisum = jax.ops.segment_sum(si,batch_index,nsys)
                l0 = sisum*(qtot - Qtot)
                q0 = q0 - si*l0[batch_index]
                x0 = jnp.concatenate((l0,q0))
            else:
                x0 = None

            solver = self.solver.lower()
            if solver == "bicg":
                x = jax.scipy.sparse.linalg.bicgstab(matvec, b,x0=x0)[0]
            elif solver == "gmres":
                x = jax.scipy.sparse.linalg.gmres(matvec, b,x0=x0)[0]
            elif solver == "cg":
                print("Warning: Use of cg solver for Qeq is not recommended")
                x = jax.scipy.sparse.linalg.cg(matvec, b,x0=x0)[0]
            else:
                raise NotImplementedError(f"solver '{solver}' is not implemented. Choose one of [bicg, gmres]")


            q = x[nsys:]
            q_ = jax.lax.stop_gradient(q)

        eself = 0.5 * Aii * q_**2 + chi * q_

        phi = jax.ops.segment_sum(Aij * q_[edge_dst], edge_src, species.shape[0])
        if do_recip:
            erec, _ = ewald_reciprocal(q_, *ewald_params)
        epair = 0.5 * q_ * phi

        if self.charges_key in inputs:
            if self.c3_key is not None:
                c3 = c3 + inputs[self.c3_key]
            if self.c4_key is not None:
                c4 = c4 + inputs[self.c4_key]
            eself = eself + c3 * q_**3 + c4 * q_**4
            training = "training" in inputs.get("flags", {})
            if self.trainable and training:
                Aii_ = jax.lax.stop_gradient(Aii)
                chi_ = jax.lax.stop_gradient(chi)
                phi_ = jax.lax.stop_gradient(phi)
                c3_ = jax.lax.stop_gradient(c3)
                c4_ = jax.lax.stop_gradient(c4)
                switch_ = jax.lax.stop_gradient(switch)
                Aij_ = jax.lax.stop_gradient(Aij)
                phi_ = jax.ops.segment_sum(
                    Aij_ * q[edge_dst], edge_src, species.shape[0]
                )

                dedq = Aii_ * q + chi_ + phi_ + 3 * c3_ * q**2 + 4 * c4_ * q**3
                dedq = jax.ops.segment_sum(
                    switch_ * (dedq[edge_src] - dedq[edge_dst]) ** 2,
                    edge_src,
                    species.shape[0],
                )
                etrain = (
                    0.5 * Aii_ * q**2
                    + chi_ * q
                    + 0.5 * q * phi_
                    + c3_ * q**3
                    + c4_ * q**4
                )
                if do_recip:
                    etrain = etrain + erec

        energy = eself + epair
        if do_recip:
            energy = energy + erec

        energy_key = self.energy_key if self.energy_key is not None else self.name
        energy_unit = au.get_multiplier(self._energy_unit)
        output = {
            **inputs,
            self.charges_key: q,
            energy_key: energy*energy_unit,
        }
        if do_recip:
            output[energy_key + "_reciprocal"] = erec*energy_unit

        training = "training" in inputs.get("flags", {})
        if self.charges_key in inputs and self.trainable and training:
            output[energy_key + "_regularization"] = regularization
            output[energy_key + "_dedq"] = dedq*energy_unit
            output[energy_key + "_etrain"] = etrain*energy_unit
        return output


class ChargeCorrection(nn.Module):
    """Charge correction scheme
    
    FID: CHARGE_CORRECTION

    Used to correct the provided charges to sum to the total charge of the system.
    """
    key: str = "charges"
    """Key for the charges in the inputs"""
    output_key: str = None
    """Key for the corrected charges in the outputs. If None, it is the same as the input key"""
    dq_key: str = "delta_qtot"
    """Key for the deviation of the raw charge sum in the outputs"""
    ratioeta_key: str = None
    """Key for the ratio of hardness between AIM and free atom in the inputs. Used to adjust charge redistribution."""
    trainable: bool = False
    """Whether the parameters are trainable"""
    cn_key: str = None
    """Key for the coordination number in the inputs. Used to adjust charge redistribution."""
    total_charge_key: str = "total_charge"
    """Key for the total charge in the inputs"""
    _energy_unit: str = "Ha"
    """The energy unit of the model. **Automatically set by FENNIX**"""

    FID: ClassVar[str] = "CHARGE_CORRECTION"

    @nn.compact
    def __call__(self, inputs) -> Any:
        species = inputs["species"]
        batch_index = inputs["batch_index"]
        nsys = inputs["natoms"].shape[0]
        q = inputs[self.key]
        if q.shape[-1] == 1:
            q = jnp.squeeze(q, axis=-1)
        qtot = jax.ops.segment_sum(q, batch_index, nsys)
        Qtot = (
            inputs[self.total_charge_key].astype(q.dtype)
            if self.total_charge_key in inputs
            else jnp.zeros(qtot.shape[0], dtype=q.dtype)
        )
        dq = Qtot - qtot

        Jii = D3_HARDNESSES
        ai = D3_VDW_RADII
        eta = [jii + (2.0 / np.pi) ** 0.5 / aii for jii, aii in zip(Jii, ai)]
        if self.trainable:
            eta = jnp.abs(self.param("eta", lambda key: jnp.asarray(eta)))[species]
        else:
            eta = jnp.asarray(eta)[species]

        if self.ratioeta_key is not None:
            ratioeta = inputs[self.ratioeta_key]
            if ratioeta.shape[-1] == 1:
                ratioeta = jnp.squeeze(ratioeta, axis=-1)
            eta = eta * ratioeta

        s = (1.0e-6 + 2 * jnp.abs(eta)) ** (-1)
        if self.cn_key is not None:
            cn = inputs[self.cn_key]
            if cn.shape[-1] == 1:
                cn = jnp.squeeze(cn, axis=-1)
            s = s * cn

        f = dq / jax.ops.segment_sum(s, batch_index, nsys)

        qf = q + s * f[batch_index]

        energy_unit = au.get_multiplier(self._energy_unit)
        ecorr = (0.5*energy_unit) * eta * (qf - q) ** 2
        output_key = self.output_key if self.output_key is not None else self.key
        return {
            **inputs,
            output_key: qf,
            self.dq_key: dq,
            "charge_correction_energy": ecorr,
        }

class DistributeElectrons(nn.Module):
    """Distribute valence electrons between the atoms

    FID: DISTRIBUTE_ELECTRONS

    Used to predict charges that sum to the total charge of the system.
    """
    embedding_key: str
    """Key for the embedding in the inputs that is used to predict an 'electron affinity' weight"""
    output_key: Union[str,None] = None
    """Key for the charges in the outputs"""
    total_charge_key: str = "total_charge"
    """Key for the total charge in the inputs"""

    FID: ClassVar[str] = "DISTRIBUTE_ELECTRONS"

    @nn.compact
    def __call__(self, inputs) -> Any:
        species = inputs["species"]
        Nel = jnp.asarray(VALENCE_ELECTRONS)[species]

        ei = nn.Dense(1, use_bias=True, name="wi")(inputs[self.embedding_key]).squeeze(-1)
        wi = jax.nn.softplus(ei)

        batch_index = inputs["batch_index"]
        nsys = inputs["natoms"].shape[0]
        wtot = jax.ops.segment_sum(wi, inputs["batch_index"], inputs["natoms"].shape[0])

        Qtot = (
            inputs[self.total_charge_key].astype(ei.dtype)
            if self.total_charge_key in inputs
            else jnp.zeros(nsys, dtype=ei.dtype)
        )
        Neltot = jax.ops.segment_sum(Nel, batch_index, nsys) - Qtot

        f = Neltot / wtot
        Ni = wi* f[batch_index]
        q = Nel-Ni
        

        output_key = self.output_key if self.output_key is not None else self.name
        return {
            **inputs,
            output_key: q,
        }




#!/usr/bin/env python3
"""Implementation of SchNet embedding.

Done by Côme Cattin, 2024.
"""

import dataclasses
from typing import Callable, Dict, Sequence, Union, ClassVar

import flax.linen as nn
import jax

from ...utils.activations import ssp
from ..misc.encodings import RadialBasis, SpeciesEncoding
from ..misc.nets import FullyConnectedNet


class SchNetEmbedding(nn.Module):
    """SchNet embedding.

    Continuous filter convolutional neural network for modeling quantum
    interactions.

    ### References
    SCHÜTT, Kristof, KINDERMANS, Pieter-Jan, SAUCEDA FELIX, Huziel Enoc, et al.
    Schnet: A continuous-filter convolutional neural network for
    modeling quantum interactions.
    Advances in neural information processing systems, 2017, vol. 30.
    https://proceedings.neurips.cc/paper_files/paper/2017/file/303ed4c69846ab36c2904d3ba8573050-Paper.pdf

    Parameters
    ----------
    dim : int, default=64
        The dimension of the embedding.
    nlayers : int, default=3
        The number of interaction layers.
    graph_key : str, default="graph"
        The key for the graph input.
    embedding_key : str, default="embedding"
        The key for the embedding output.
    radial_basis : dict, default={}
        The radial basis function parameters.
    species_encoding : dict, default={}
        The species encoding parameters.
    activation : Union[Callable, str], default=ssp
        The activation function.
    """

    _graphs_properties: Dict
    dim: int = 64
    """The dimension of the embedding."""
    nlayers: int = 3
    """The number of interaction layers."""
    conv_hidden: Sequence[int] = dataclasses.field(
        default_factory=lambda: [64, 64]
    )
    """The hidden layers for the edge network."""
    graph_key: str = "graph"
    """The key for the graph input."""
    embedding_key: str = "embedding"
    """The key for the embedding output."""
    radial_basis: dict = dataclasses.field(default_factory=dict)
    """The radial basis function parameters. See `fennol.models.misc.encodings.RadialBasis`."""
    species_encoding: dict = dataclasses.field(default_factory=dict)
    """The species encoding parameters. See `fennol.models.misc.encodings.SpeciesEncoding`."""
    activation: Union[Callable, str] = "ssp"
    """The activation function."""

    FID: ClassVar[str] = "SCHNET"

    @nn.compact
    def __call__(self, inputs):
        """Forward pass."""
        species = inputs["species"]
        graph = inputs[self.graph_key]
        switch = graph["switch"][:, None]
        edge_src, edge_dst = graph["edge_src"], graph["edge_dst"]
        cutoff = self._graphs_properties[self.graph_key]["cutoff"]
        onehot = SpeciesEncoding(
            **self.species_encoding, name="SpeciesEncoding"
        )(species)

        xi_prev_layer = nn.Dense(
            self.dim, name="species_linear", use_bias=False
        )(onehot)

        distances = graph["distances"]
        radial_basis = RadialBasis(
            **{
                "end": cutoff,
                **self.radial_basis,
                "name": "RadialBasis",
            }
        )(distances)

        def atom_wise(xi, i, layer):
            return nn.Dense(
                self.dim, name=f"atom_wise_{i}_{layer}", use_bias=True
            )(xi)

        # Interaction layer
        for layer in range(self.nlayers):
            # Atom-wise
            xi = atom_wise(xi_prev_layer, 1, layer)

            # cfconv
            w_l = FullyConnectedNet(
                [*self.conv_hidden, self.dim],
                activation=self.activation,
                name=f"filter_weight_{layer}",
                use_bias=True,
            )(radial_basis)
            xi_j = xi[edge_dst]
            xi = jax.ops.segment_sum(
                self.activation(w_l) * xi_j * switch, edge_src, species.shape[0]
            )

            # Atom-wise
            xi = atom_wise(xi, 2, layer)

            # Activation
            xi = self.activation(xi)

            # Atom-wise
            xi = atom_wise(xi, 3, layer)

            # Residual connection
            xi = xi + xi_prev_layer
            xi_prev_layer = xi

        output = {
            **inputs,
            self.embedding_key: xi,
        }
        return output


if __name__ == "__main__":
    pass



from numpy import pi


class AtomicUnits:
    EV = 27.211386024367243  # Hartree to eV
    KCALPERMOL = 627.5096080305927  # Hartree to kcal/mol
    KJPERMOL = 2625.5002  # Hartree to kJ/mol
    BOHR = 0.52917721  # Bohr to Angstrom
    MPROT = 1836.1526734252586  # proton mass
    HBAR = 1.0  # Planck's constant
    FS = 2.4188843e-2  # AU time to femtoseconds
    PS = FS / 1000  # AU time to picoseconds
    KELVIN = 3.15774e5  # Hartree to Kelvin
    THZ = 1000.0 / FS  # AU frequency to THz
    NNEWTON = 82.387  # AU Force to nNewton
    CM1 = 219471.52  # Hartree to cm-1
    GMOL_AFS = BOHR / (MPROT * FS)  # AU momentum to (g/mol).A/fs
    KBAR = 294210.2648438959  # Hartree/bohr**3 to kbar
    ATM = KBAR * 1000.0 / 1.01325  # Hartree/bohr**3 to atm
    GPA = 0.1 * KBAR  # Hartree/bohr**3 to GPa
    DEBYE = 2.541746  # e.Bohr to Debye
    FSC = 1.0 / 137.035999084  # Fine structure constant
    MOL = 6.02214129e+23
    KCAL = KCALPERMOL * MOL
    KJ = KJPERMOL * MOL
    RY = 2.  # Hartree to Rydberg

    mapping = {
        "1": 1.0,
        "AU": 1.0,
        "HA":1.0,
        "EV": EV,
        "MEV": 1.0e3 * EV, # milli electronvolt
        "KCALPERMOL": KCALPERMOL,
        "KJPERMOL": KJPERMOL,
        "ANGSTROM": BOHR,
        "BOHR": 1.0 / BOHR,
        "AMU": 1.0 / MPROT,
        "FEMTOSECONDS": FS,
        "FS": FS,
        "PICOSECONDS": FS / 1000.0,
        "PS": FS / 1000.0,
        "KELVIN": KELVIN,
        "K": KELVIN,
        "THZ": THZ,
        "TRADHZ": THZ / (2 * pi),
        "CM-1": CM1,
        "CM1": CM1,
        "GMOLAFS": GMOL_AFS,
        "KBAR": KBAR,
        "ATM": ATM,
        "GPA": GPA,
        "DEBYE": DEBYE,
        "FSC": FSC,
        "MOL": MOL,
        "MOLE": MOL,
        "KCAL": KCAL,
        "KJ": KJ,
        "RY": RY,
    }

    @staticmethod
    def get_multiplier(unit_string):
        unit_string = unit_string.upper().strip()

        multiplier = 1.0

        unit_start = 0
        unit_stop = 0
        in_power = False
        power_start = 0
        power_stop = 0
        tmp_power = 1.0
        for i in range(len(unit_string)):
            current_char = unit_string[i]
            # print(current_char)
            if current_char == "^":
                if in_power:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                in_power = True
                unit_stop = i - 1
                if unit_stop - unit_start < 0:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

            elif current_char == "{":
                if not in_power:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                if i + 1 >= len(unit_string):
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                power_start = i + 1

            elif current_char == "}":
                if not in_power:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                in_power = False
                power_stop = i - 1

                if power_stop - power_start < 0:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError
                else:
                    tmp_power_read = int(unit_string[power_start : power_stop + 1])
                    tmp_power = tmp_power * tmp_power_read

                unit_substring = unit_string[unit_start : unit_stop + 1]
                tmp_unit = AtomicUnits.mapping[unit_substring]

                multiplier = multiplier * (tmp_unit**tmp_power)

                power_start = 0
                power_stop = 0
                unit_start = i + 1
                unit_stop = 0

            elif current_char == "*":
                if in_power:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                if unit_start == i:
                    unit_start = i + 1
                    tmp_power = 1.0
                    continue

                unit_stop = i - 1
                if unit_stop - unit_start < 0:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                unit_substring = unit_string[unit_start : unit_stop + 1]
                tmp_unit = AtomicUnits.mapping[unit_substring]
                multiplier = multiplier * (tmp_unit**tmp_power)

                unit_start = i + 1
                unit_stop = 0
                tmp_power = 1.0

            elif current_char == "/":
                if in_power:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                if unit_start == i:
                    unit_start = i + 1
                    tmp_power = -1.0
                    continue

                unit_stop = i - 1
                if unit_stop - unit_start < 0:
                    print("Error: Syntax error in unit '" + unit_string + "' !")
                    raise ValueError

                unit_substring = unit_string[unit_start : unit_stop + 1]
                tmp_unit = AtomicUnits.mapping[unit_substring]
                multiplier = multiplier * (tmp_unit**tmp_power)

                unit_start = i + 1
                unit_stop = 0
                tmp_power = -1.0

            else:
                if i + 1 >= len(unit_string):
                    if in_power:
                        print("Error: Syntax error in unit '" + unit_string + "' !")
                        raise ValueError

                    unit_stop = i
                    if unit_stop - unit_start < 0:
                        print("Error: Syntax error in unit '" + unit_string + "' !")
                        raise ValueError

                    unit_substring = unit_string[unit_start : unit_stop + 1]
                    tmp_unit = AtomicUnits.mapping[unit_substring]
                    multiplier = multiplier * (tmp_unit**tmp_power)

        return multiplier
